/*
	Copyright (C) 2012-2030

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

//#pragma once
#ifndef __X64ENCODER_H__
#define __X64ENCODER_H__


#include "x64Instructions.h"



#define INVALIDCODEBLOCK		0xffffffff

// some x64 instructions have more than 1 opcode
#define MAKE2OPCODE(opcode1,opcode2)			( ( ( opcode2 ) << 8 ) | ( opcode1 ) )
#define MAKE3OPCODE(opcode1,opcode2,opcode3)	( ( ( opcode3 ) << 16 ) | ( ( opcode2 ) << 8 ) | ( opcode1 ) )
#define MAKE4OPCODE(opcode1,opcode2,opcode3,opcode4)	( ( ( opcode4 ) << 24 ) | ( ( opcode3 ) << 16 ) | ( ( opcode2 ) << 8 ) | ( opcode1 ) )
#define GETOPCODE1(multiopcode)					( ( multiopcode ) & 0xff )
#define GETOPCODE2(multiopcode)					( ( ( multiopcode ) >> 8 ) & 0xff )
#define GETOPCODE3(multiopcode)					( ( ( multiopcode ) >> 16 ) & 0xff )
#define GETOPCODE4(multiopcode)					( ( ( multiopcode ) >> 24 ) & 0xff )


// defines for vpcmp
enum { VPCMP_EQ = 0, VPCMP_LT = 1, VPCMP_LE = 2, VPCMP_FALSE = 3, VPCMP_NE = 4, VPCMP_GE = 5, VPCMP_GT = 6, VPCMP_TRUE = 7 };


// this not only encodes from a source cpu to a target cpu but also holds the code and runs it
// lNumberOfCodeBlocks must be a power of 2
// the total amount allocated to the x64 code area is lCodeBlockSize*lNumberOfCodeBlocks*2
// lCodeBlockSize is in bytes

class x64Encoder
{

public:
	
#ifdef ENABLE_RECOMPILER_HASH_TABLE
	// need to know what code is where
	// this contains the full source address (no, I'll use address/4 or address/8) and the index of the x64 block in the array of x64 code
	// you look up by using the lower bits of the source address
	long* x64CodeHashTable;
#endif
	
#ifdef ENABLE_RECOMPILER_NEXT_ADDRESS
	// we also need to know the source cpu address that points to next instruction after block - not next instruction to execute, just the next one
	long* x64CodeSourceNextAddress;
#endif

	// this is going to point to the current code area
	char* x64CodeArea;
	
	// this is going to be where all the live cached x64 code is at - no need for this after testing
	char* LiveCodeArea;
	
#ifdef ENABLE_ALTERNATE_STREAM
	// I'll need a work area before copying all dynamic code into actual code block
	char* AlternateStream;
#endif

	long lCodeBlockSize_PowerOfTwo, lCodeBlockSize, lCodeBlockSize_Mask;
	long lNumberOfCodeBlocks;
	
	// I need the index for the current code block
	long x64CurrentCodeBlockIndex;
	
	// the total amount of space allocated for code area
	long x64TotalSpaceAllocated;

	// need to know the address where to start putting the next x64 instruction
	long x64NextOffset;
	
	// need to know the address of the start of the current x64 instruction being encoded
	long x64CurrentStartOffset;
	
	// need to know offset first x64 byte of current source cpu instruction being encoded
	long x64CurrentSourceCpuInstructionStartOffset;
	
	// need to know the Source Cpu Address that is currently being worked on
	long x64CurrentSourceAddress;
	
	// need to keep track of the source cpu instruction size
	long x64SourceCpuInstructionSize;

	// need to know if we are currently encoding or if we're ready to execute
	bool isEncoding;
	
	// we also ned to know that it was set up successfully
	bool isReadyForUse;
	
	// the size of code in bytes in alternate stream
	long lAlternateStreamSize;
	
	// holds the code for alternate stream
	unsigned long long ullAlternateStream;

	// no need for this stuff after testing - modern cpus can handle this stuff without any tricks needed
	// we can get more control over pre-loading of data by separating pre-allocation and instructions into different streams
	// I don't know the effect on performance, so I'll add in some way to control operation of this
//	long PreAllocationBitmap;	// just set the bits for registers, then if a bit is already set copy the streams (PreAlloc then Inst), reset, repeat
//	long PreAllocationStreamSize;
//	long InstructionStreamSize;
//	char* PreAllocationStream;
//	char* InstructionStream;
	
	// constructor
	// CountOfCodeBlocks must be a power of 2, OR ELSE!!!!
	x64Encoder ( long SizePerCodeBlock_PowerOfTwo, long CountOfCodeBlocks );
	
	// destructor
	~x64Encoder ( void );
	
	// flush current code block
	bool FlushCurrentCodeBlock ( void );
	
	// flush dynamic code block - must do this once after creation before you can use it
	bool FlushCodeBlock ( long IndexOfCodeBlock );
	
	// flush instruction cache line for current process - have to use this with dynamic code or else
	bool FlushICache ( long long baseAddress, long NumberOfBytes );

	inline void Emit_AltStreamToBlock8 ( long lCodeBlockIndex )
	{
		((unsigned long long*)LiveCodeArea) [ ( lCodeBlockIndex << lCodeBlockSize_PowerOfTwo ) >> 3 ] = ullAlternateStream;
	}


	inline void Emit_AltStreamToBlock4 ( long lCodeBlockIndex )
	{
		((unsigned long*)LiveCodeArea) [ ( lCodeBlockIndex << lCodeBlockSize_PowerOfTwo ) >> 2 ] = ullAlternateStream;
	}

	
	// get/set current instruction offset
	long GetCurrentInstructionOffset ( void );
	void SetCurrentInstructionOffset ( long offset );

	// get the size of a code block for encoder - should help with determining if there is more space available
	long GetCodeBlockSize ( void );
	
	// get the size of the current dynamic code being worked out - should help with determining if there is more space available
	// this will only work when currently working in the alternate stream
	long GetAlternateStreamCurrentSize ( void );
	
	// start writing code to alternate stream
	void SwitchToAlternateStream ( void );
	
	// start writing code to live code area
	void SwitchToLiveStream ( void );
	
	// copy dynamic code from alternate stream to current position in live stream
	void CopyToLiveStream ( void );

	// call this before beginning an x64 code block
	bool StartCodeBlock ( long lCodeBlockIndex );

	// call this when you are done with the code block
	bool EndCodeBlock ( void );

	// gets a pointer to the start of current code block
	char* Get_CodeBlock_StartPtr ();
	
	// gets a pointer to the end of current code block (where there SHOULD be return instruction)
	char* Get_CodeBlock_EndPtr ();

	// get the start pointer for an arbitrary code block number
	char* Get_XCodeBlock_StartPtr ( long lCodeBlockIndex );
	
	// get the current pointer in code block
	char* Get_CodeBlock_CurrentPtr ();

	// call this before starting to encode a single instruction from source cpu
	bool StartInstructionBlock ( void );

	// call this when done encoding a single instructin from source cpu
	bool EndInstructionBlock ( void );

	// takes you to the start of the current source cpu instruction you were encoding - cuz you may need to backtrack
	bool UndoInstructionBlock ( void );
	
	// this returns the amount of space remaining in the current code block
	long x64CurrentCodeBlockSpaceRemaining ( void );
	
	// invalidate a code block that has been overwritten in source cpu memory or is otherwise no longer valid
	bool x64InvalidateCodeBlock ( long lSourceAddress );

	// check if code is already encoded and ready to run
	bool x64IsEncodedAndReady ( long lSourceAddress );

	// executes a code block and returns address of the next instruction
	//long long ExecuteCodeBlock ( long lSourceAddress );
	inline long long ExecuteCodeBlock ( long lSourceAddress )
	{
		// function will return the address of the next instruction to execute
		typedef long long (*asm_function) ( void );
		
		volatile asm_function x64Function;

		//long lCodeBlockIndex;
		
		// get the index for the code block to put x64 code in
		//lCodeBlockIndex = ( lSourceAddress & ( lNumberOfCodeBlocks - 1 ) );
		
		//x64Function = (asm_function) ( & ( x64CodeArea [ lCodeBlockIndex * lNumberOfCodeBlocks ] ) );
		x64Function = (asm_function) ( & ( x64CodeArea [ lSourceAddress << lCodeBlockSize_PowerOfTwo ] ) );

		return x64Function ();
	}



	// **** Functions For Encoding of x64 Instructions **** //
	
	// encode an instruction with no register arguments
	bool x64Encode16 ( long x64InstOpcode );
	bool x64Encode32 ( long x64InstOpcode );
	bool x64Encode64 ( long x64InstOpcode );
	
	// encode a move immediate instruction using fewest bytes possible
	bool x64EncodeMovImm64 ( long x64DestRegIndex, long long Immediate64 );
	bool x64EncodeMovImm32 ( long x64DestRegIndex, long Immediate32 );
	bool x64EncodeMovImm16 ( long x64DestRegIndex, short Immediate16 );
	bool x64EncodeMovImm8 ( long x64DestRegIndex, long Immediate8 );

	// encode a single register x64 instruction into code block
	bool x64EncodeReg32 ( long x64InstOpcode, long ModRMOpcode, long x64Reg );
	bool x64EncodeReg64 ( long x64InstOpcode, long ModRMOpcode, long x64Reg );
	
	// encode a register-register x64 instruction into code block
	bool x64EncodeRegReg16 ( long x64InstOpcode, long x64DestReg_Reg_Opcode, long x64SourceReg_RM_Base );
	bool x64EncodeRegReg32 ( long x64InstOpcode, long x64DestReg_Reg_Opcode, long x64SourceReg_RM_Base );
	bool x64EncodeRegReg64 ( long x64InstOpcode, long x64DestReg_Reg_Opcode, long x64SourceReg_RM_Base );

	// encode a x64 Imm8 instruction into code block
	bool x64EncodeReg16Imm8 ( long x64InstOpcode, long ModRMOpcode, long x64DestReg_RM_Base, char cImmediate8 );
	bool x64EncodeReg32Imm8 ( long x64InstOpcode, long ModRMOpcode, long x64DestReg_RM_Base, char cImmediate8 );
	bool x64EncodeReg64Imm8 ( long x64InstOpcode, long ModRMOpcode, long x64DestReg_RM_Base, char cImmediate8 );
	
	// encode x64 Imm8 instruction but with memory
	bool x64EncodeMem16Imm8 ( long x64InstOpcode, long ModRMOpcode, long BaseAddressReg, long IndexReg, long Scale, long Offset, char cImmediate8 );
	bool x64EncodeMem32Imm8 ( long x64InstOpcode, long ModRMOpcode, long BaseAddressReg, long IndexReg, long Scale, long Offset, char cImmediate8 );
	bool x64EncodeMem64Imm8 ( long x64InstOpcode, long ModRMOpcode, long BaseAddressReg, long IndexReg, long Scale, long Offset, char cImmediate8 );

	
	// encode 16-bit immediate x64 instruction
	bool x64EncodeReg16Imm16 ( long x64InstOpcode, long ModRMOpcode, long x64DestReg, short cImmediate16 );

	// encode a reg-immediate 32 bit x64 instruction into code block
	bool x64EncodeReg32Imm32 ( long x64InstOpcode, long ModRMOpcode, long x64DestReg, long cImmediate32 );

	// encode a reg-immediate 64 bit x64 instruction into code block
	bool x64EncodeReg64Imm32 ( long x64InstOpcode, long ModRMOpcode, long x64DestReg, long cImmediate32 );

	// encode for operation on accumulator
	bool x64EncodeAcc16Imm16 ( long x64InstOpcode, short cImmediate16 );
	bool x64EncodeAcc32Imm32 ( long x64InstOpcode, long cImmediate32 );
	bool x64EncodeAcc64Imm32 ( long x64InstOpcode, long cImmediate32 );

	
	
	// *** These functions work with memory accesses ***
	// ** NOTE ** Do not use sp/esp/rsp or bp/ebp/rbp with these functions until I read the x64 docs more thoroughly!!!

	// encode an x64 instruction that addresses memory
	bool x64EncodeRegMem16 ( long x64InstOpcode, long x64DestReg, long BaseAddressReg, long IndexReg, long Scale, long Offset );
	bool x64EncodeRegMem32 ( long x64InstOpcode, long x64DestReg, long BaseAddressReg, long IndexReg, long Scale, long Offset );
	bool x64EncodeRegMem64 ( long x64InstOpcode, long x64DestReg, long BaseAddressReg, long IndexReg, long Scale, long Offset );

	bool x64EncodeMemImm8 ( long x64InstOpcode, long Mod, char Imm8, long BaseAddressReg, long IndexReg, long Scale, long Offset );
	bool x64EncodeMemImm16 ( long x64InstOpcode, long Mod, short Imm16, long BaseAddressReg, long IndexReg, long Scale, long Offset );
	bool x64EncodeMemImm32 ( long x64InstOpcode, long Mod, long Imm32, long BaseAddressReg, long IndexReg, long Scale, long Offset );
	bool x64EncodeMemImm64 ( long x64InstOpcode, long Mod, long Imm32, long BaseAddressReg, long IndexReg, long Scale, long Offset );
	
	// these are for encoding just immediate values
	bool x64EncodeImmediate8 ( char Imm8 );
	bool x64EncodeImmediate16 ( short Imm16 );
	bool x64EncodeImmediate32 ( long Imm32 );
	bool x64EncodeImmediate64 ( long long Imm64 );

	// encode a return instruction
	bool x64EncodeReturn ( void );
	
	
	
	// *** testing *** encode RIP-offset addressing
	bool x64EncodeRipOffset16 ( long x64InstOpcode, long x64Reg, char* DataAddress, bool bIsSourceReg = false );
	bool x64EncodeRipOffset32 ( long x64InstOpcode, long x64Reg, char* DataAddress, bool bIsSourceReg = false );
	bool x64EncodeRipOffset64 ( long x64InstOpcode, long x64Reg, char* DataAddress, bool bIsSourceReg = false );
	bool x64EncodeRipOffsetImm8 ( long x64InstOpcode, long x64Reg, char* DataAddress, char Imm8, bool bIsSourceReg = false );
	bool x64EncodeRipOffsetImm16 ( long x64InstOpcode, long x64Reg, char* DataAddress, short Imm16, bool bIsSourceReg = false );
	bool x64EncodeRipOffsetImm32 ( long x64InstOpcode, long x64Reg, char* DataAddress, long Imm32, bool bIsSourceReg = false );
	bool x64EncodeRipOffsetImm64 ( long x64InstOpcode, long x64Reg, char* DataAddress, long Imm32, bool bIsSourceReg = false );
	
	bool x64EncodeRipOffset16Imm8 ( long x64InstOpcode, long x64Reg, char* DataAddress, char Imm8, bool bIsSourceReg = false );
	bool x64EncodeRipOffset32Imm8 ( long x64InstOpcode, long x64Reg, char* DataAddress, char Imm8, bool bIsSourceReg = false );
	bool x64EncodeRipOffset64Imm8 ( long x64InstOpcode, long x64Reg, char* DataAddress, char Imm8, bool bIsSourceReg = false );
	
	
	
	// ** Encoding vector/sse instructions ** //

	// encode an avx instruction (64-bit version) with an immediate byte on the end
	bool x64EncodeRegVImm8 ( long L, long w, long pp, long mmmmm, long avxInstOpcode, long REG_R, long vvvv, long RM_B, char cImmediate8 );
	bool x64EncodeRegMemVImm8 ( long L, long w, long pp, long mmmmm, long avxInstOpcode, long REG_R_Dest, long vvvv, long x64RM_B_Base, long x64IndexReg, long Scale, long Offset, char Imm8 );
	
	// encode avx instruction for a single 32-bit load store to/from memory
	bool x64EncodeRegMem32S ( long pp, long mmmmm, long avxInstOpcode, long avxDestSrcReg, long avxBaseReg, long avxIndexReg, long Scale, long Offset );
	bool x64EncodeRegMemV ( long L, long w, long pp, long mmmmm, long avxInstOpcode, long REG_R_Dest, long vvvv, long x64RM_B_Base, long x64IndexReg, long Scale, long Offset );
	bool x64EncodeRegMem256 ( long pp, long mmmmm, long avxInstOpcode, long avxDestSrcReg, long avxBaseReg, long avxIndexReg, long Scale, long Offset );

	// encode an avx-512 instruction
	bool x64EncodeRegRegEV(long L, long w, long pp, long mm, long avxInstOpcode, long REG_R, long vvvv, long RM_B, long z, long k, long rnd_enable = 0, long rnd_type = 0);
	bool x64EncodeRegRegEVImm8(long L, long w, long pp, long mm, long avxInstOpcode, long REG_R, long vvvv, long RM_B, char cImmediate8, long z, long k);

	// encode an avx instruction (64-bit version) that is just register-register
	bool x64EncodeRegRegV ( long L, long w, long pp, long mmmmm, long avxInstOpcode, long REG_R, long vvvv, long RM_B );

	bool x64EncodeRipOffsetV ( long L, long w, long pp, long mmmmm, long avxInstOpcode, long REG_R, long vvvv, char* DataAddress );
	bool x64EncodeRipOffsetVImm8(long L, long w, long pp, long mmmmm, long avxInstOpcode, long REG_R, long vvvv, char* DataAddress, char Imm8);

	// avx512
	bool x64EncodeRipOffsetEV(long L, long w, long pp, long mmm, long avxInstOpcode, long REG_R, long vvvv, char* DataAddress, long z, long k, long broadcast = 0);
	bool x64EncodeRipOffsetEVImm8(long L, long w, long pp, long mmm, long avxInstOpcode, long REG_R, long vvvv, char* DataAddress, char Imm8, long z, long k, long broadcast = 0);

	// **** x64 Instuctions **** //
	
	// ** General Purpose Register Instructions ** //
	
	
	// * jump/branch instructions * //
	
	static const int c_iNumOfBranchOffsets = 32768;
	static const int c_iNumOfBranchOffsets_Mask = c_iNumOfBranchOffsets - 1;
	
//	long JmpOffset [ 256 ];
	long BranchOffset [ c_iNumOfBranchOffsets ];
/*
	long JmpOffsetEndCount;
+++++	long JmpOffset_End [ 512 ];

	long BranchOffsetEndCount;
	long BranchOffset_End [ 512 ];
*/

	// jump to address in 64-bit register
	bool JmpReg64 ( long x64Reg );
	bool JmpMem64 ( long long* SrcPtr );

	// these are used to make a short term hop while encoding x64 instructions
	bool Jmp ( long Offset, unsigned long Label );
	bool Jmp_NE ( long Offset, unsigned long Label );
	bool Jmp_E ( long Offset, unsigned long Label );
	bool Jmp_L ( long Offset, unsigned long Label );
	bool Jmp_LE ( long Offset, unsigned long Label );
	bool Jmp_G ( long Offset, unsigned long Label );
	bool Jmp_GE ( long Offset, unsigned long Label );
	bool Jmp_B ( long Offset, unsigned long Label );
	bool Jmp_BE ( long Offset, unsigned long Label );
	bool Jmp_A ( long Offset, unsigned long Label );
	bool Jmp_AE ( long Offset, unsigned long Label );
	bool Jmp_S(long Offset, unsigned long Label);
	bool Jmp_NS(long Offset, unsigned long Label);

	
	// jump short
	bool Jmp8 ( char Offset, unsigned long Label );
	
	// jump short if equal
	bool Jmp8_E ( char Offset, unsigned long Label );
	
	// jump short if not equal
	bool Jmp8_NE ( char Offset, unsigned long Label );
	
	// jump short if unsigned above (carry flag=0 and zero flag=0)
	bool Jmp8_A ( char Offset, unsigned long Label );

	// jump short if unsigned above or equal (carry flag=0)
	bool Jmp8_AE ( char Offset, unsigned long Label );
	
	// jump short if unsigned below (carry flag=1)
	bool Jmp8_B ( char Offset, unsigned long Label );

	// jump short if unsigned below or equal (carry flag=1 or zero flag=1)
	bool Jmp8_BE ( char Offset, unsigned long Label );

	bool Jmp8_G ( char Offset, unsigned long Label );
	bool Jmp8_GE ( char Offset, unsigned long Label );
	bool Jmp8_L ( char Offset, unsigned long Label );
	bool Jmp8_LE ( char Offset, unsigned long Label );
	
	// jump short if not overflow
	bool Jmp8_NO ( char Offset, unsigned long Label );

	// jump if CX/ECX/RCX is zero
	bool Jmp8_CXZ ( char Offset, unsigned long Label );
	bool Jmp8_ECXZ ( char Offset, unsigned long Label );
	bool Jmp8_RCXZ ( char Offset, unsigned long Label );

	

	// these are used to set the target address of jump once you find out what it is
	bool SetJmpTarget ( unsigned long Label );
	
	// set jump target address for a short jump
	bool SetJmpTarget8 ( unsigned long Label );
	
	// self-optimizing unconditional jump
	bool JMP ( void* Target );
	
	// self-optimizing jump on equal
	bool JMP_E ( void* Target );
	
	// self-optimizing jump on not-equal
	bool JMP_NE ( void* Target );

	bool JMP_LE(void* Target);
	bool JMP_G(void* Target);
	bool JMP_L(void* Target);
	bool JMP_GE(void* Target);
	bool JMP_B(void* Target);
	bool JMP_AE(void* Target);

	
	// absolute jump with 32-bit rip-offset
	bool Jmp ( void* Target );
	bool Jmp_B ( void* Target );
	bool Jmp_O(void* Target);

	
	// absolute call with 32-bit rip-offset
	bool Call ( const void* Target );

//	bool SetJmpTarget_End ( void );

	// mov
	bool MovRegReg16 ( long DestReg, long SrcReg );
	bool MovRegReg32 ( long DestReg, long SrcReg );
	bool MovRegReg64 ( long DestReg, long SrcReg );
	bool MovRegImm8 ( long DestReg, char Imm8 );
	bool MovRegImm16 ( long DestReg, short Imm16 );
	bool MovRegImm32 ( long DestReg, long Imm32 );
	bool MovRegImm64 ( long DestReg, long long Imm64 );
	bool MovReg64Imm32 ( long DestReg, long Imm32 );
	bool MovRegToMem8 ( long SrcReg, long AddressReg, long IndexReg, long Scale, long Offset );
	bool MovRegFromMem8 ( long DestReg, long AddressReg, long IndexReg, long Scale, long Offset );
	bool MovRegToMem16 ( long SrcReg, long AddressReg, long IndexReg, long Scale, long Offset );
	bool MovRegFromMem16 ( long DestReg, long AddressReg, long IndexReg, long Scale, long Offset );
	bool MovRegToMem32 ( long SrcReg, long AddressReg, long IndexReg, long Scale, long Offset );
	bool MovRegFromMem32 ( long DestReg, long AddressReg, long IndexReg, long Scale, long Offset );
	bool MovRegToMem64 ( long SrcReg, long AddressReg, long IndexReg, long Scale, long Offset );
	bool MovRegFromMem64 ( long DestReg, long AddressReg, long IndexReg, long Scale, long Offset );
	
	bool MovMemImm8 ( char Imm8, long AddressReg, long IndexReg, long Scale, long Offset );
	bool MovMemImm16 ( short Imm16, long AddressReg, long IndexReg, long Scale, long Offset );
	bool MovMemImm32 ( long Imm32, long AddressReg, long IndexReg, long Scale, long Offset );
	bool MovMemImm64 ( long Imm32, long AddressReg, long IndexReg, long Scale, long Offset );

	// self-optimizing instructions
	bool MovReg16ImmX ( long DestReg, short Imm16 );
	bool MovReg32ImmX ( long DestReg, long Imm32 );
	bool MovReg64ImmX ( long DestReg, long long Imm64 );

	
	// *** testing ***
	bool MovRegToMem8 ( char* Address, long SrcReg );
	bool MovRegFromMem8 ( long DestReg, char* Address );
	bool MovRegToMem16 ( short* Address, long SrcReg );
	bool MovRegFromMem16 ( long DestReg, short* Address );
	bool MovRegToMem32 ( long* Address, long SrcReg );
	bool MovRegFromMem32 ( long DestReg, long* Address );
	bool MovRegToMem64 ( long long* Address, long SrcReg );
	bool MovRegFromMem64 ( long DestReg, long long* Address );
	
	inline bool MovMemReg8 ( char* Address, long SrcReg ) { return MovRegToMem8 ( Address, SrcReg ); }
	inline bool MovRegMem8 ( long DestReg, char* Address ) { return MovRegFromMem8 ( DestReg, Address ); }
	inline bool MovMemReg16 ( short* Address, long SrcReg ) { return MovRegToMem16 ( Address, SrcReg ); }
	inline bool MovRegMem16 ( long DestReg, short* Address ) { return MovRegFromMem16 ( DestReg, Address ); }
	inline bool MovMemReg32 ( long* Address, long SrcReg ) { return MovRegToMem32 ( Address, SrcReg ); }
	inline bool MovRegMem32 ( long DestReg, long* Address ) { return MovRegFromMem32 ( DestReg, Address ); }
	inline bool MovMemReg64 ( long long* Address, long SrcReg ) { return MovRegToMem64 ( Address, SrcReg ); }
	inline bool MovRegMem64 ( long DestReg, long long* Address ) { return MovRegFromMem64 ( DestReg, Address ); }

	bool MovMemImm8 ( char* DestPtr, char Imm8 );
	bool MovMemImm16 ( short* DestPtr, short Imm16 );
	bool MovMemImm32 ( long* DestPtr, long Imm32 );
	bool MovMemImm64 ( long long* DestPtr, long Imm32 );

	// load variables from memory //
	inline bool LoadMem8 ( long DstReg, void* MemoryAddress ) { return MovRegFromMem8 ( DstReg, (char*) MemoryAddress ); }
	inline bool LoadMem16 ( long DstReg, void* MemoryAddress ) { return MovRegFromMem16 ( DstReg, (short*) MemoryAddress ); }
	inline bool LoadMem32 ( long DstReg, void* MemoryAddress ) { return MovRegFromMem32 ( DstReg, (long*) MemoryAddress ); }
	inline bool LoadMem64 ( long DstReg, void* MemoryAddress ) { return MovRegFromMem64 ( DstReg, (long long*) MemoryAddress ); }
	
	// store variables to memory //
	inline bool StoreMem8 ( void* MemoryAddress, long SrcReg ) { return MovRegToMem8 ( (char*) MemoryAddress, SrcReg ); }
	inline bool StoreMem16 ( void* MemoryAddress, long SrcReg ) { return MovRegToMem16 ( (short*) MemoryAddress, SrcReg ); }
	inline bool StoreMem32 ( void* MemoryAddress, long SrcReg ) { return MovRegToMem32 ( (long*) MemoryAddress, SrcReg ); }
	inline bool StoreMem64 ( void* MemoryAddress, long SrcReg ) { return MovRegToMem64 ( (long long*) MemoryAddress, SrcReg ); }
	
	// set register to constant //
	//inline bool LoadImm8 ( long DstReg, char Imm8 ) { return MovRegImm8 ( DstReg, Imm8 ); }
	inline bool LoadImm16 ( long DstReg, short Imm16 ) { return MovRegImm16 ( DstReg, Imm16 ); }
	inline bool LoadImm32 ( long DstReg, long Imm32 ) { return MovRegImm32 ( DstReg, Imm32 ); }
	inline bool LoadImm64 ( long DstReg, long long Imm64 ) { return MovRegImm64 ( DstReg, Imm64 ); }

	// set memory to constant //
	//inline bool StoreImm8 ( void* MemoryAddress, char Imm8 ) { return MovMemImm8 ( (char*) MemoryAddress, Imm8 ); }
	inline bool StoreImm16 ( void* MemoryAddress, short Imm16 ) { return MovMemImm16 ( (short*) MemoryAddress, Imm16 ); }
	inline bool StoreImm32 ( void* MemoryAddress, long Imm32 ) { return MovMemImm32 ( (long*) MemoryAddress, Imm32 ); }
	inline bool StoreImm64 ( void* MemoryAddress, long Imm32 ) { return MovMemImm64 ( (long long*) MemoryAddress, Imm32 ); }
	
	
//	bool MovRegToMem32S ( long SrcReg, long AddressReg, long IndexReg, long Scale, long Offset );
//	bool MovRegFromMem32S ( long DestReg, long AddressReg, long IndexReg, long Scale, long Offset );

	// add
	bool AddRegReg16 ( long DestReg, long SrcReg );
	bool AddRegReg32 ( long DestReg, long SrcReg );
	bool AddRegReg64 ( long DestReg, long SrcReg );
	bool AddRegImm16 ( long DestReg, short Imm16 );
	bool AddRegImm32 ( long DestReg, long Imm32 );
	bool AddRegImm64 ( long DestReg, long Imm32 );
	
	bool AddRegMem16 ( long DestReg, long AddressReg, long IndexReg, long Scale, long Offset );
	bool AddRegMem32 ( long DestReg, long AddressReg, long IndexReg, long Scale, long Offset );
	bool AddRegMem64 ( long DestReg, long AddressReg, long IndexReg, long Scale, long Offset );
	bool AddMemReg16 ( long SrcReg, long AddressReg, long IndexReg, long Scale, long Offset );
	bool AddMemReg32 ( long SrcReg, long AddressReg, long IndexReg, long Scale, long Offset );
	bool AddMemReg64 ( long SrcReg, long AddressReg, long IndexReg, long Scale, long Offset );
	bool AddMemImm16 ( short Imm16, long AddressReg, long IndexReg, long Scale, long Offset );
	bool AddMemImm32 ( long Imm32, long AddressReg, long IndexReg, long Scale, long Offset );
	bool AddMemImm64 ( long Imm32, long AddressReg, long IndexReg, long Scale, long Offset );

	bool AddRegMem16 ( long DestReg, short* SrcPtr );
	bool AddRegMem32 ( long DestReg, long* SrcPtr );
	bool AddRegMem64 ( long DestReg, long long* SrcPtr );
	bool AddMemReg16 ( short* DestPtr, long SrcReg );
	bool AddMemReg32 ( long* DestPtr, long SrcReg );
	bool AddMemReg64 ( long long* DestPtr, long SrcReg );
	bool AddMemImm16 ( short* DestPtr, short Imm16 );
	bool AddMemImm32 ( long* DestPtr, long Imm32 );
	bool AddMemImm64 ( long long* DestPtr, long Imm32 );
	

	// 8-bit immediate
	bool AddReg16Imm8 ( long DestReg, char Imm8 );
	bool AddReg32Imm8 ( long DestReg, char Imm8 );
	bool AddReg64Imm8 ( long DestReg, char Imm8 );

	bool AddMem16Imm8 ( short* DestPtr, char Imm8 );
	bool AddMem32Imm8 ( long* DestPtr, char Imm8 );
	bool AddMem64Imm8 ( long long* DestPtr, char Imm8 );
	
	// add with accumulator
	bool AddAcc16Imm16 ( short Imm16 );
	bool AddAcc32Imm32 ( long Imm32 );
	bool AddAcc64Imm32 ( long Imm32 );
	
	// self-optimizing versions
	bool AddReg16ImmX ( long DestReg, short ImmX );
	bool AddReg32ImmX ( long DestReg, long ImmX );
	bool AddReg64ImmX ( long DestReg, long ImmX );
	
	// self-optimizing
	bool AddMem16ImmX ( short* DestPtr, short ImmX );
	bool AddMem32ImmX ( long* DestPtr, long ImmX );
	bool AddMem64ImmX ( long long* DestPtr, long ImmX );
	
	
	// adc - add with carry
	
	bool AdcRegReg16 ( long DestReg, long SrcReg );
	bool AdcRegReg32 ( long DestReg, long SrcReg );
	bool AdcRegReg64 ( long DestReg, long SrcReg );
	bool AdcRegImm16 ( long DestReg, short Imm16 );
	bool AdcRegImm32 ( long DestReg, long Imm32 );
	bool AdcRegImm64 ( long DestReg, long Imm32 );
	
	bool AdcRegMem16 ( long DestReg, long AddressReg, long IndexReg, long Scale, long Offset );
	bool AdcRegMem32 ( long DestReg, long AddressReg, long IndexReg, long Scale, long Offset );
	bool AdcRegMem64 ( long DestReg, long AddressReg, long IndexReg, long Scale, long Offset );
	bool AdcMemReg16 ( long SrcReg, long AddressReg, long IndexReg, long Scale, long Offset );
	bool AdcMemReg32 ( long SrcReg, long AddressReg, long IndexReg, long Scale, long Offset );
	bool AdcMemReg64 ( long SrcReg, long AddressReg, long IndexReg, long Scale, long Offset );
	bool AdcMemImm16 ( short Imm16, long AddressReg, long IndexReg, long Scale, long Offset );
	bool AdcMemImm32 ( long Imm32, long AddressReg, long IndexReg, long Scale, long Offset );
	bool AdcMemImm64 ( long Imm32, long AddressReg, long IndexReg, long Scale, long Offset );

	bool AdcRegMem16 ( long DestReg, short* SrcPtr );
	bool AdcRegMem32 ( long DestReg, long* SrcPtr );
	bool AdcRegMem64 ( long DestReg, long long* SrcPtr );
	bool AdcMemReg16 ( short* DestPtr, long SrcReg );
	bool AdcMemReg32 ( long* DestPtr, long SrcReg );
	bool AdcMemReg64 ( long long* DestPtr, long SrcReg );
	bool AdcMemImm16 ( short* DestPtr, short Imm16 );
	bool AdcMemImm32 ( long* DestPtr, long Imm32 );
	bool AdcMemImm64 ( long long* DestPtr, long Imm32 );
	

	// 8-bit immediate
	bool AdcReg16Imm8 ( long DestReg, char Imm8 );
	bool AdcReg32Imm8 ( long DestReg, char Imm8 );
	bool AdcReg64Imm8 ( long DestReg, char Imm8 );

	bool AdcMem16Imm8 ( short* DestPtr, char Imm8 );
	bool AdcMem32Imm8 ( long* DestPtr, char Imm8 );
	bool AdcMem64Imm8 ( long long* DestPtr, char Imm8 );
	
	// add with accumulator
	bool AdcAcc16Imm16 ( short Imm16 );
	bool AdcAcc32Imm32 ( long Imm32 );
	bool AdcAcc64Imm32 ( long Imm32 );
	
	// self-optimizing versions
	bool AdcReg16ImmX ( long DestReg, short ImmX );
	bool AdcReg32ImmX ( long DestReg, long ImmX );
	bool AdcReg64ImmX ( long DestReg, long ImmX );
	
	// self-optimizing
	bool AdcMem16ImmX ( short* DestPtr, short ImmX );
	bool AdcMem32ImmX ( long* DestPtr, long ImmX );
	bool AdcMem64ImmX ( long long* DestPtr, long ImmX );
	
	
	// and
	bool AndRegReg16 ( long DestReg, long SrcReg );
	bool AndRegReg32 ( long DestReg, long SrcReg );
	bool AndRegReg64 ( long DestReg, long SrcReg );
	bool AndRegImm16 ( long DestReg, short Imm16 );
	bool AndRegImm32 ( long DestReg, long Imm32 );
	bool AndRegImm64 ( long DestReg, long Imm32 );

	bool AndRegMem16 ( long DestReg, long AddressReg, long IndexReg, long Scale, long Offset );
	bool AndRegMem32 ( long DestReg, long AddressReg, long IndexReg, long Scale, long Offset );
	bool AndRegMem64 ( long DestReg, long AddressReg, long IndexReg, long Scale, long Offset );
	bool AndMemReg16 ( long SrcReg, long AddressReg, long IndexReg, long Scale, long Offset );
	bool AndMemReg32 ( long SrcReg, long AddressReg, long IndexReg, long Scale, long Offset );
	bool AndMemReg64 ( long SrcReg, long AddressReg, long IndexReg, long Scale, long Offset );
	bool AndMemImm16 ( short Imm16, long AddressReg, long IndexReg, long Scale, long Offset );
	bool AndMemImm32 ( long Imm32, long AddressReg, long IndexReg, long Scale, long Offset );
	bool AndMemImm64 ( long Imm32, long AddressReg, long IndexReg, long Scale, long Offset );
	
	bool AndRegMem16 ( long DestReg, short* SrcPtr );
	bool AndRegMem32 ( long DestReg, long* SrcPtr );
	bool AndRegMem64 ( long DestReg, long long* SrcPtr );
	bool AndMemReg16 ( short* DestPtr, long SrcReg );
	bool AndMemReg32 ( long* DestPtr, long SrcReg );
	bool AndMemReg64 ( long long* DestPtr, long SrcReg );
	bool AndMemImm16 ( short* DestPtr, short Imm16 );
	bool AndMemImm32 ( long* DestPtr, long Imm32 );
	bool AndMemImm64 ( long long* DestPtr, long Imm32 );

	// 8-bit immediate
	bool AndReg16Imm8 ( long DestReg, char Imm8 );
	bool AndReg32Imm8 ( long DestReg, char Imm8 );
	bool AndReg64Imm8 ( long DestReg, char Imm8 );

	bool AndMem16Imm8 ( short* DestPtr, char Imm8 );
	bool AndMem32Imm8 ( long* DestPtr, char Imm8 );
	bool AndMem64Imm8 ( long long* DestPtr, char Imm8 );
	
	// add with accumulator
	bool AndAcc16Imm16 ( short Imm16 );
	bool AndAcc32Imm32 ( long Imm32 );
	bool AndAcc64Imm32 ( long Imm32 );
	
	// self-optimizing versions
	bool AndReg16ImmX ( long DestReg, short ImmX );
	bool AndReg32ImmX ( long DestReg, long ImmX );
	bool AndReg64ImmX ( long DestReg, long ImmX );
	
	// self-optimizing
	bool AndMem16ImmX ( short* DestPtr, short ImmX );
	bool AndMem32ImmX ( long* DestPtr, long ImmX );
	bool AndMem64ImmX ( long long* DestPtr, long ImmX );
	
	
	// bsr - bit scan reverse - leading zero count
	bool BsrRegReg16 ( long DestReg, long SrcReg );
	bool BsrRegReg32 ( long DestReg, long SrcReg );
	bool BsrRegReg64 ( long DestReg, long SrcReg );
	
	bool BsrRegMem16 ( long DestReg, short* SrcPtr );
	bool BsrRegMem32 ( long DestReg, long* SrcPtr );
	bool BsrRegMem64 ( long DestReg, long long* SrcPtr );
	
	
	// bt - bit test - stores specified bit of value in carry flag
	bool BtRegReg16 ( long Reg, long BitSelectReg );
	bool BtRegReg32 ( long Reg, long BitSelectReg );
	bool BtRegReg64 ( long Reg, long BitSelectReg );
	bool BtRegImm16 ( long Reg, char Imm8 );
	bool BtRegImm32 ( long Reg, char Imm8 );
	bool BtRegImm64 ( long Reg, char Imm8 );
	
	// with memory
	bool BtMemImm16 ( short* DestPtr, char Imm8 );
	bool BtMemImm32 ( long* DestPtr, char Imm8 );
	bool BtMemImm64 ( long long* DestPtr, char Imm8 );
	bool BtMemReg16 ( long BitSelectReg, long AddressReg, long IndexReg, long Scale, long Offset );
	bool BtMemReg32 ( long BitSelectReg, long AddressReg, long IndexReg, long Scale, long Offset );
	bool BtMemReg64 ( long BitSelectReg, long AddressReg, long IndexReg, long Scale, long Offset );
	bool BtMemImm16 ( char Imm8, long AddressReg, long IndexReg, long Scale, long Offset );
	bool BtMemImm32 ( char Imm8, long AddressReg, long IndexReg, long Scale, long Offset );
	bool BtMemImm64 ( char Imm8, long AddressReg, long IndexReg, long Scale, long Offset );

	// btc - bit test and compliment
	bool BtcRegReg16 ( long Reg, long BitSelectReg );
	bool BtcRegReg32 ( long Reg, long BitSelectReg );
	bool BtcRegReg64 ( long Reg, long BitSelectReg );
	bool BtcRegImm16 ( long Reg, char Imm8 );
	bool BtcRegImm32 ( long Reg, char Imm8 );
	bool BtcRegImm64 ( long Reg, char Imm8 );
	
	// with memory
	bool BtcMemReg16 ( long BitSelectReg, long AddressReg, long IndexReg, long Scale, long Offset );
	bool BtcMemReg32 ( long BitSelectReg, long AddressReg, long IndexReg, long Scale, long Offset );
	bool BtcMemReg64 ( long BitSelectReg, long AddressReg, long IndexReg, long Scale, long Offset );
	bool BtcMemImm16 ( char Imm8, long AddressReg, long IndexReg, long Scale, long Offset );
	bool BtcMemImm32 ( char Imm8, long AddressReg, long IndexReg, long Scale, long Offset );
	bool BtcMemImm64 ( char Imm8, long AddressReg, long IndexReg, long Scale, long Offset );

	bool BtcMemImm16(short* DestPtr, char Imm8);
	bool BtcMemImm32(long* DestPtr, char Imm8);
	bool BtcMemImm64(long long* DestPtr, char Imm8);

	// btr - bit test and reset
	bool BtrRegReg16 ( long Reg, long BitSelectReg );
	bool BtrRegReg32 ( long Reg, long BitSelectReg );
	bool BtrRegReg64 ( long Reg, long BitSelectReg );
	bool BtrRegImm16 ( long Reg, char Imm8 );
	bool BtrRegImm32 ( long Reg, char Imm8 );
	bool BtrRegImm64 ( long Reg, char Imm8 );
	
	// with memory
	bool BtrMemReg16 ( long BitSelectReg, long AddressReg, long IndexReg, long Scale, long Offset );
	bool BtrMemReg32 ( long BitSelectReg, long AddressReg, long IndexReg, long Scale, long Offset );
	bool BtrMemReg64 ( long BitSelectReg, long AddressReg, long IndexReg, long Scale, long Offset );
	bool BtrMemImm16 ( char Imm8, long AddressReg, long IndexReg, long Scale, long Offset );
	bool BtrMemImm32 ( char Imm8, long AddressReg, long IndexReg, long Scale, long Offset );
	bool BtrMemImm64 ( char Imm8, long AddressReg, long IndexReg, long Scale, long Offset );

	bool BtrMem32Imm ( long* DestPtr, char Imm8 );
	
	// bts - bit test and set
	bool BtsRegReg16 ( long Reg, long BitSelectReg );
	bool BtsRegReg32 ( long Reg, long BitSelectReg );
	bool BtsRegReg64 ( long Reg, long BitSelectReg );
	bool BtsRegImm16 ( long Reg, char Imm8 );
	bool BtsRegImm32 ( long Reg, char Imm8 );
	bool BtsRegImm64 ( long Reg, char Imm8 );

	// with memory
	bool BtsMemReg16 ( long BitSelectReg, long AddressReg, long IndexReg, long Scale, long Offset );
	bool BtsMemReg32 ( long BitSelectReg, long AddressReg, long IndexReg, long Scale, long Offset );
	bool BtsMemReg64 ( long BitSelectReg, long AddressReg, long IndexReg, long Scale, long Offset );
	bool BtsMemImm16 ( char Imm8, long AddressReg, long IndexReg, long Scale, long Offset );
	bool BtsMemImm32 ( char Imm8, long AddressReg, long IndexReg, long Scale, long Offset );
	bool BtsMemImm64 ( char Imm8, long AddressReg, long IndexReg, long Scale, long Offset );
	
	bool BtsMemImm32(long* DestPtr, char Imm8);

	// cbw, cwde, cdqe - sign extention for RAX - from byte to word, word to double word, or double word to quadword respectively
	bool Cbw ( void );
	bool Cwde ( void );
	bool Cdqe ( void );
	
	// cwd, cdq, cqo - sign extension RAX into RAX:RDX
	bool Cwd ( void );
	bool Cdq ( void );
	bool Cqo ( void );

	// cmov - conditional mov
	bool CmovERegReg16 ( long DestReg, long SrcReg );
	bool CmovNERegReg16 ( long DestReg, long SrcReg );
	bool CmovBRegReg16 ( long DestReg, long SrcReg );
	bool CmovBERegReg16 ( long DestReg, long SrcReg );
	bool CmovARegReg16 ( long DestReg, long SrcReg );
	bool CmovAERegReg16 ( long DestReg, long SrcReg );
	bool CmovORegReg16(long DestReg, long SrcReg);
	bool CmovERegReg32 ( long DestReg, long SrcReg );
	bool CmovNERegReg32 ( long DestReg, long SrcReg );
	bool CmovBRegReg32 ( long DestReg, long SrcReg );
	bool CmovBERegReg32 ( long DestReg, long SrcReg );
	bool CmovARegReg32 ( long DestReg, long SrcReg );
	bool CmovAERegReg32 ( long DestReg, long SrcReg );
	bool CmovORegReg32(long DestReg, long SrcReg);
	bool CmovERegReg64 ( long DestReg, long SrcReg );
	bool CmovNERegReg64 ( long DestReg, long SrcReg );
	bool CmovBRegReg64 ( long DestReg, long SrcReg );
	bool CmovBERegReg64 ( long DestReg, long SrcReg );
	bool CmovARegReg64 ( long DestReg, long SrcReg );
	bool CmovAERegReg64 ( long DestReg, long SrcReg );
	bool CmovORegReg64(long DestReg, long SrcReg);

	
	// with memory
	bool CmovERegMem16 ( long DestReg, long AddressReg, long IndexReg, long Scale, long Offset );
	bool CmovNERegMem16 ( long DestReg, long AddressReg, long IndexReg, long Scale, long Offset );
	bool CmovERegMem32 ( long DestReg, long AddressReg, long IndexReg, long Scale, long Offset );
	bool CmovNERegMem32 ( long DestReg, long AddressReg, long IndexReg, long Scale, long Offset );
	bool CmovERegMem64 ( long DestReg, long AddressReg, long IndexReg, long Scale, long Offset );
	bool CmovNERegMem64 ( long DestReg, long AddressReg, long IndexReg, long Scale, long Offset );
	bool CmovBRegMem64 ( long DestReg, long AddressReg, long IndexReg, long Scale, long Offset );

	bool CmovAERegMem16 ( long DestReg, long AddressReg, long IndexReg, long Scale, long Offset );
	bool CmovAERegMem32 ( long DestReg, long AddressReg, long IndexReg, long Scale, long Offset );
	bool CmovAERegMem64 ( long DestReg, long AddressReg, long IndexReg, long Scale, long Offset );

	bool CmovERegMem16 ( long DestReg, short* Mem16 );
	bool CmovNERegMem16 ( long DestReg, short* Mem16 );
	bool CmovBRegMem16 ( long DestReg, short* Mem16 );
	bool CmovBERegMem16 ( long DestReg, short* Mem16 );
	bool CmovARegMem16 ( long DestReg, short* Mem16 );
	bool CmovAERegMem16 ( long DestReg, short* Mem16 );
	bool CmovERegMem32 ( long DestReg, long* Mem32 );
	bool CmovNERegMem32 ( long DestReg, long* Mem32 );
	bool CmovBRegMem32 ( long DestReg, long* Mem32 );
	bool CmovBERegMem32 ( long DestReg, long* Mem32 );
	bool CmovARegMem32 ( long DestReg, long* Mem32 );
	bool CmovAERegMem32 ( long DestReg, long* Mem32 );
	bool CmovERegMem64 ( long DestReg, long long* Mem64 );
	bool CmovNERegMem64 ( long DestReg, long long* Mem64 );
	bool CmovBRegMem64 ( long DestReg, long long* Mem64 );
	bool CmovBERegMem64 ( long DestReg, long long* Mem64 );
	bool CmovARegMem64 ( long DestReg, long long* Mem64 );
	bool CmovAERegMem64 ( long DestReg, long long* Mem64 );

	
	// signed conditional mov
	
	bool CmovLRegReg16 ( long DestReg, long SrcReg );
	bool CmovLERegReg16 ( long DestReg, long SrcReg );
	bool CmovGRegReg16 ( long DestReg, long SrcReg );
	bool CmovGERegReg16 ( long DestReg, long SrcReg );
	bool CmovLRegReg32 ( long DestReg, long SrcReg );
	bool CmovLERegReg32 ( long DestReg, long SrcReg );
	bool CmovGRegReg32 ( long DestReg, long SrcReg );
	bool CmovGERegReg32 ( long DestReg, long SrcReg );
	bool CmovLRegReg64 ( long DestReg, long SrcReg );
	bool CmovLERegReg64 ( long DestReg, long SrcReg );
	bool CmovGRegReg64 ( long DestReg, long SrcReg );
	bool CmovGERegReg64 ( long DestReg, long SrcReg );

	bool CmovSRegReg16 ( long DestReg, long SrcReg );
	bool CmovSRegReg32 ( long DestReg, long SrcReg );
	bool CmovSRegReg64 ( long DestReg, long SrcReg );
	
	bool CmovSRegMem16 ( long DestReg, short* Mem16 );
	bool CmovSRegMem32 ( long DestReg, long* Mem32 );
	bool CmovSRegMem64 ( long DestReg, long long* Mem64 );

	bool CmovNSRegReg16 ( long DestReg, long SrcReg );
	bool CmovNSRegReg32 ( long DestReg, long SrcReg );
	bool CmovNSRegReg64 ( long DestReg, long SrcReg );
	
	bool CmovNSRegMem16 ( long DestReg, short* Mem16 );
	bool CmovNSRegMem32 ( long DestReg, long* Mem32 );
	bool CmovNSRegMem64 ( long DestReg, long long* Mem64 );
	

	// cmp - compare two values
	bool CmpRegReg16 ( long SrcReg1, long SrcReg2 );
	bool CmpRegReg32 ( long SrcReg1, long SrcReg2 );
	bool CmpRegReg64 ( long SrcReg1, long SrcReg2 );
	bool CmpRegImm16 ( long SrcReg, short Imm16 );
	bool CmpRegImm32 ( long SrcReg, long Imm32 );
	bool CmpRegImm64 ( long SrcReg, long Imm32 );

	bool CmpRegMem8 ( long SrcReg, long AddressReg, long IndexReg, long Scale, long Offset );
	bool CmpRegMem16 ( long SrcReg, long AddressReg, long IndexReg, long Scale, long Offset );
	bool CmpRegMem32 ( long SrcReg, long AddressReg, long IndexReg, long Scale, long Offset );
	bool CmpRegMem64 ( long SrcReg, long AddressReg, long IndexReg, long Scale, long Offset );
	bool CmpMemReg16 ( long SrcReg, long AddressReg, long IndexReg, long Scale, long Offset );
	bool CmpMemReg32 ( long SrcReg, long AddressReg, long IndexReg, long Scale, long Offset );
	bool CmpMemReg64 ( long SrcReg, long AddressReg, long IndexReg, long Scale, long Offset );
	bool CmpMemImm16 ( short Imm16, long AddressReg, long IndexReg, long Scale, long Offset );
	bool CmpMemImm32 ( long Imm32, long AddressReg, long IndexReg, long Scale, long Offset );
	bool CmpMemImm64 ( long Imm32, long AddressReg, long IndexReg, long Scale, long Offset );

	bool CmpRegMem16 ( long DestReg, short* SrcPtr );
	bool CmpRegMem32 ( long DestReg, long* SrcPtr );
	bool CmpRegMem64 ( long DestReg, long long* SrcPtr );
	bool CmpMemReg16 ( short* DestPtr, long SrcReg );
	bool CmpMemReg32 ( long* DestPtr, long SrcReg );
	bool CmpMemReg64 ( long long* DestPtr, long SrcReg );
	bool CmpMemImm8 ( char* DestPtr, char Imm8 );
	bool CmpMemImm16 ( short* DestPtr, short Imm16 );
	bool CmpMemImm32 ( long* DestPtr, long Imm32 );
	bool CmpMemImm64 ( long long* DestPtr, long Imm32 );

	// 8-bit immediate
	bool CmpReg16Imm8 ( long DestReg, char Imm8 );
	bool CmpReg32Imm8 ( long DestReg, char Imm8 );
	bool CmpReg64Imm8 ( long DestReg, char Imm8 );

	bool CmpMem16Imm8 ( short* DestPtr, char Imm8 );
	bool CmpMem32Imm8 ( long* DestPtr, char Imm8 );
	bool CmpMem64Imm8 ( long long* DestPtr, char Imm8 );
	
	// add with accumulator
	bool CmpAcc16Imm16 ( short Imm16 );
	bool CmpAcc32Imm32 ( long Imm32 );
	bool CmpAcc64Imm32 ( long Imm32 );
	
	// self-optimizing versions
	bool CmpReg16ImmX ( long DestReg, short ImmX );
	bool CmpReg32ImmX ( long DestReg, long ImmX );
	bool CmpReg64ImmX ( long DestReg, long ImmX );
	
	// self-optimizing
	bool CmpMem16ImmX ( short* DestPtr, short ImmX );
	bool CmpMem32ImmX ( long* DestPtr, long ImmX );
	bool CmpMem64ImmX ( long long* DestPtr, long ImmX );
	
	
	// div - unsigned divide - divides d:a by register, puts quotient in a, remainder in d
	bool DivRegReg16 ( long SrcReg );
	bool DivRegReg32 ( long SrcReg );
	bool DivRegReg64 ( long SrcReg );

	bool DivRegMem16 ( short* SrcPtr );
	bool DivRegMem32 ( long* SrcPtr );
	bool DivRegMem64 ( long long* SrcPtr );

	bool DivRegMem16 ( long AddressReg, long IndexReg, long Scale, long Offset );
	bool DivRegMem32 ( long AddressReg, long IndexReg, long Scale, long Offset );
	bool DivRegMem64 ( long AddressReg, long IndexReg, long Scale, long Offset );

	// idiv - signed divide
	bool IdivRegReg16 ( long SrcReg );
	bool IdivRegReg32 ( long SrcReg );
	bool IdivRegReg64 ( long SrcReg );

	bool IdivRegMem16 ( short* SrcPtr );
	bool IdivRegMem32 ( long* SrcPtr );
	bool IdivRegMem64 ( long long* SrcPtr );

	bool IdivRegMem16 ( long AddressReg, long IndexReg, long Scale, long Offset );
	bool IdivRegMem32 ( long AddressReg, long IndexReg, long Scale, long Offset );
	bool IdivRegMem64 ( long AddressReg, long IndexReg, long Scale, long Offset );
	
	// imul - signed multiply
	bool ImulRegReg16 ( long SrcReg );
	bool ImulRegReg32 ( long SrcReg );
	bool ImulRegReg64 ( long SrcReg );

	bool ImulRegMem16 ( short* SrcPtr );
	bool ImulRegMem32 ( long* SrcPtr );
	bool ImulRegMem64 ( long long* SrcPtr );

	bool ImulRegMem16 ( long AddressReg, long IndexReg, long Scale, long Offset );
	bool ImulRegMem32 ( long AddressReg, long IndexReg, long Scale, long Offset );
	bool ImulRegMem64 ( long AddressReg, long IndexReg, long Scale, long Offset );

	// lea - can be used as 3 operand add instruction
	bool LeaRegRegReg16 ( long DestReg, long Src1Reg, long Src2Reg );
	bool LeaRegRegReg32 ( long DestReg, long Src1Reg, long Src2Reg );
	bool LeaRegRegReg64 ( long DestReg, long Src1Reg, long Src2Reg );
	bool LeaRegRegImm16 ( long DestReg, long SrcReg, long Imm16 );
	bool LeaRegRegImm32 ( long DestReg, long SrcReg, long Imm32 );
	bool LeaRegRegImm64 ( long DestReg, long SrcReg, long Imm32 );

	// loads the address into register using a 4-byte offset
	//bool LeaRegMem32 ( long DestReg, void* SrcPtr );
	bool LeaRegMem64 ( long DestReg, void* SrcPtr );

	
	// movsx - move with sign extension
	bool MovsxReg16Reg8 ( long DestReg, long SrcReg );
	bool MovsxReg32Reg8 ( long DestReg, long SrcReg );
	bool MovsxReg64Reg8 ( long DestReg, long SrcReg );

	bool MovsxReg16Mem8 ( long DestReg, char* SrcPtr );
	bool MovsxReg32Mem8 ( long DestReg, char* SrcPtr );
	bool MovsxReg64Mem8 ( long DestReg, char* SrcPtr );

	bool MovsxReg16Mem8 ( long DestReg, long AddressReg, long IndexReg, long Scale, long Offset );
	bool MovsxReg32Mem8 ( long DestReg, long AddressReg, long IndexReg, long Scale, long Offset );
	bool MovsxReg64Mem8 ( long DestReg, long AddressReg, long IndexReg, long Scale, long Offset );

	bool MovsxReg32Reg16 ( long DestReg, long SrcReg );
	bool MovsxReg64Reg16 ( long DestReg, long SrcReg );
	
	bool MovsxReg32Mem16 ( long DestReg, short* SrcPtr );
	bool MovsxReg64Mem16 ( long DestReg, short* SrcPtr );
	
	
	bool MovsxdReg64Reg32 ( long DestReg, long SrcReg );

	bool MovsxdReg64Mem32 ( long DestReg, long* SrcPtr );
	
	bool MovsxReg32Mem16 ( long DestReg, long AddressReg, long IndexReg, long Scale, long Offset );
	bool MovsxReg64Mem16 ( long DestReg, long AddressReg, long IndexReg, long Scale, long Offset );
	bool MovsxdReg64Mem32 ( long DestReg, long AddressReg, long IndexReg, long Scale, long Offset );

	
	// movzx - move with zero extension
	bool MovzxReg16Reg8 ( long DestReg, long SrcReg );
	bool MovzxReg32Reg8 ( long DestReg, long SrcReg );
	bool MovzxReg64Reg8 ( long DestReg, long SrcReg );

	bool MovzxReg16Mem8 ( long DestReg, char* SrcPtr );
	bool MovzxReg32Mem8 ( long DestReg, char* SrcPtr );
	bool MovzxReg64Mem8 ( long DestReg, char* SrcPtr );

	bool MovzxReg16Mem8 ( long DestReg, long AddressReg, long IndexReg, long Scale, long Offset );
	bool MovzxReg32Mem8 ( long DestReg, long AddressReg, long IndexReg, long Scale, long Offset );
	bool MovzxReg64Mem8 ( long DestReg, long AddressReg, long IndexReg, long Scale, long Offset );

	bool MovzxReg32Reg16 ( long DestReg, long SrcReg );
	bool MovzxReg64Reg16 ( long DestReg, long SrcReg );

	bool MovzxReg32Mem16 ( long DestReg, short* SrcPtr );
	bool MovzxReg64Mem16 ( long DestReg, short* SrcPtr );

	bool MovzxReg32Mem16 ( long DestReg, long AddressReg, long IndexReg, long Scale, long Offset );
	bool MovzxReg64Mem16 ( long DestReg, long AddressReg, long IndexReg, long Scale, long Offset );
	
	
	// mul - unsigned multiply
	bool MulRegReg16 ( long SrcReg );
	bool MulRegReg32 ( long SrcReg );
	bool MulRegReg64 ( long SrcReg );
	
	bool MulRegMem16 ( short* SrcPtr );
	bool MulRegMem32 ( long* SrcPtr );
	bool MulRegMem64 ( long long* SrcPtr );
	
	bool MulRegMem16 ( long AddressReg, long IndexReg, long Scale, long Offset );
	bool MulRegMem32 ( long AddressReg, long IndexReg, long Scale, long Offset );
	bool MulRegMem64 ( long AddressReg, long IndexReg, long Scale, long Offset );

	// neg
	bool NegReg16 ( long DestReg );
	bool NegReg32 ( long DestReg );
	bool NegReg64 ( long DestReg );
	
	bool NegMem16 ( long AddressReg, long IndexReg, long Scale, long Offset );
	bool NegMem32 ( long AddressReg, long IndexReg, long Scale, long Offset );
	bool NegMem64 ( long AddressReg, long IndexReg, long Scale, long Offset );

	bool NegMem16 ( short* DestPtr );
	bool NegMem32 ( long* DestPtr );
	bool NegMem64 ( long long* DestPtr );

	// not 
	bool NotReg16 ( long DestReg );
	bool NotReg32 ( long DestReg );
	bool NotReg64 ( long DestReg );

	bool NotMem16 ( long AddressReg, long IndexReg, long Scale, long Offset );
	bool NotMem32 ( long AddressReg, long IndexReg, long Scale, long Offset );
	bool NotMem64 ( long AddressReg, long IndexReg, long Scale, long Offset );

	bool NotMem16 ( short* DestPtr );
	bool NotMem32 ( long* DestPtr );
	bool NotMem64 ( long long* DestPtr );
	
	
	// inc
	bool IncReg16 ( long DestReg );
	bool IncReg32 ( long DestReg );
	bool IncReg64 ( long DestReg );

	bool IncMem16 ( short* DestPtr );
	bool IncMem32 ( long* DestPtr );
	bool IncMem64 ( long long* DestPtr );
	
	
	// dec
	bool DecReg16 ( long DestReg );
	bool DecReg32 ( long DestReg );
	bool DecReg64 ( long DestReg );

	bool DecMem16 ( short* DestPtr );
	bool DecMem32 ( long* DestPtr );
	bool DecMem64 ( long long* DestPtr );

	
	// or
	bool OrRegReg16 ( long DestReg, long SrcReg );
	bool OrRegReg32 ( long DestReg, long SrcReg );
	bool OrRegReg64 ( long DestReg, long SrcReg );
	bool OrRegImm16 ( long DestReg, short Imm16 );
	bool OrRegImm32 ( long DestReg, long Imm32 );
	bool OrRegImm64 ( long DestReg, long Imm32 );

	bool OrRegMem16 ( long DestReg, long AddressReg, long IndexReg, long Scale, long Offset );
	bool OrRegMem32 ( long DestReg, long AddressReg, long IndexReg, long Scale, long Offset );
	bool OrRegMem64 ( long DestReg, long AddressReg, long IndexReg, long Scale, long Offset );
	bool OrMemReg16 ( long SrcReg, long AddressReg, long IndexReg, long Scale, long Offset );
	bool OrMemReg32 ( long SrcReg, long AddressReg, long IndexReg, long Scale, long Offset );
	bool OrMemReg64 ( long SrcReg, long AddressReg, long IndexReg, long Scale, long Offset );
	bool OrMemImm16 ( short Imm16, long AddressReg, long IndexReg, long Scale, long Offset );
	bool OrMemImm32 ( long Imm32, long AddressReg, long IndexReg, long Scale, long Offset );
	bool OrMemImm64 ( long Imm32, long AddressReg, long IndexReg, long Scale, long Offset );

	bool OrRegMem8(long DestReg, char* SrcPtr);
	bool OrRegMem16 ( long DestReg, short* SrcPtr );
	bool OrRegMem32 ( long DestReg, long* SrcPtr );
	bool OrRegMem64 ( long DestReg, long long* SrcPtr );
	bool OrMemReg8 ( char* DestPtr, int SrcReg );
	bool OrMemReg16 ( short* DestPtr, long SrcReg );
	bool OrMemReg32 ( long* DestPtr, long SrcReg );
	bool OrMemReg64 ( long long* DestPtr, long SrcReg );
	bool OrMemImm16 ( short* DestPtr, short Imm16 );
	bool OrMemImm32 ( long* DestPtr, long Imm32 );
	bool OrMemImm64 ( long long* DestPtr, long Imm32 );

	// 8-bit immediate
	bool OrReg16Imm8 ( long DestReg, char Imm8 );
	bool OrReg32Imm8 ( long DestReg, char Imm8 );
	bool OrReg64Imm8 ( long DestReg, char Imm8 );

	bool OrMem16Imm8 ( short* DestPtr, char Imm8 );
	bool OrMem32Imm8 ( long* DestPtr, char Imm8 );
	bool OrMem64Imm8 ( long long* DestPtr, char Imm8 );
	
	// add with accumulator
	bool OrAcc16Imm16 ( short Imm16 );
	bool OrAcc32Imm32 ( long Imm32 );
	bool OrAcc64Imm32 ( long Imm32 );
	
	// self-optimizing versions
	bool OrReg16ImmX ( long DestReg, short ImmX );
	bool OrReg32ImmX ( long DestReg, long ImmX );
	bool OrReg64ImmX ( long DestReg, long ImmX );
	
	// self-optimizing
	bool OrMem16ImmX ( short* DestPtr, short ImmX );
	bool OrMem32ImmX ( long* DestPtr, long ImmX );
	bool OrMem64ImmX ( long long* DestPtr, long ImmX );

	
	// popcnt - population count
	bool PopCnt32 ( long DestReg, long SrcReg );

	
	// pop - pop register from stack
	bool PopReg16 ( long DestReg );
	bool PopReg32 ( long DestReg );
	bool PopReg64 ( long DestReg );
	
	// push - push register onto stack
	bool PushReg16 ( long SrcReg );
	bool PushReg32 ( long SrcReg );
	bool PushReg64 ( long SrcReg );

	bool PushImm8 ( char Imm8 );
	bool PushImm16 ( short Imm16 );
	bool PushImm32 ( long Imm32 );

	// ret
	bool Ret ( void );
	
	// set - set byte instructions
	
	// unsigned versions
	
	// seta - set if unsigned above
	bool Set_A ( int DestReg );
	bool Set_A ( void* DestPtr8 );
	bool Set_A ( long AddressReg, long IndexReg, long Scale, long Offset );
	
	// setae - set if unsigned above or equal
	bool Set_AE ( int DestReg );
	bool Set_AE ( void* DestPtr8 );
	bool Set_AE ( long AddressReg, long IndexReg, long Scale, long Offset );
	
	// setb - set if unsigned below
	bool Set_B ( int DestReg );
	bool Set_B ( void* DestPtr8 );
	bool Set_B ( long AddressReg, long IndexReg, long Scale, long Offset );
	
	// setbe - set if unsigned below or equal
	bool Set_BE ( int DestReg );
	bool Set_BE ( void* DestPtr8 );
	bool Set_BE ( long AddressReg, long IndexReg, long Scale, long Offset );
	
	// sete - set if equal
	bool Set_E ( int DestReg );
	bool Set_E ( void* DestPtr8 );
	bool Set_E ( long AddressReg, long IndexReg, long Scale, long Offset );
	
	// setne - set if NOT equal
	bool Set_NE ( int DestReg );
	bool Set_NE ( void* DestPtr8 );
	bool Set_NE ( long AddressReg, long IndexReg, long Scale, long Offset );
	
	// signed versions

	// setg - set if signed greater
	bool Set_G ( int DestReg );
	bool Set_G ( void* DestPtr8 );
	bool Set_G ( long AddressReg, long IndexReg, long Scale, long Offset );

	// setge - set if signed greater
	bool Set_GE ( int DestReg );
	bool Set_GE ( void* DestPtr8 );
	bool Set_GE ( long AddressReg, long IndexReg, long Scale, long Offset );

	// setl - set if signed greater
	bool Set_L ( int DestReg );
	bool Set_L ( void* DestPtr8 );
	bool Set_L ( long AddressReg, long IndexReg, long Scale, long Offset );

	// setle - set if signed greater
	bool Set_LE ( int DestReg );
	bool Set_LE ( void* DestPtr8 );
	bool Set_LE ( long AddressReg, long IndexReg, long Scale, long Offset );

	// sets - set if signed
	bool Set_S ( int DestReg );
	bool Set_S ( void* DestPtr8 );
	bool Set_S ( long AddressReg, long IndexReg, long Scale, long Offset );
	
	// setpo - set if parity odd
	bool Set_PO ( int DestReg );
	bool Set_PO ( void* DestPtr8 );
	bool Set_PO ( long AddressReg, long IndexReg, long Scale, long Offset );
	
	
	// setb - set if unsigned below
	bool Setb ( long DestReg );
	
	// setl - set if unsigned less than
	bool Setl ( long DestReg );
	bool Setl ( long AddressReg, long IndexReg, long Scale, long Offset );


	// rcr - rotate right through carry

	// rotate right through carry once
	bool RcrReg16(long DestReg);
	bool RcrReg32(long DestReg);
	bool RcrReg64(long DestReg);


	// shl - shift left logical - these shift by the value in register c or by an immediate 8-bit value
	bool ShlRegReg16 ( long DestReg );
	bool ShlRegReg32 ( long DestReg );
	bool ShlRegReg64 ( long DestReg );
	bool ShlRegImm16 ( long DestReg, char Imm8 );
	bool ShlRegImm32 ( long DestReg, char Imm8 );
	bool ShlRegImm64 ( long DestReg, char Imm8 );
	
	bool ShlMemReg32 ( long* DestPtr );
	
	// with memory
	bool ShlMemReg16 ( long AddressReg, long IndexReg, long Scale, long Offset );
	bool ShlMemReg32 ( long AddressReg, long IndexReg, long Scale, long Offset );
	bool ShlMemReg64 ( long AddressReg, long IndexReg, long Scale, long Offset );
	bool ShlMemImm16 ( long AddressReg, long IndexReg, long Scale, long Offset, char Imm8 );
	bool ShlMemImm32 ( long AddressReg, long IndexReg, long Scale, long Offset, char Imm8 );
	bool ShlMemImm64 ( long AddressReg, long IndexReg, long Scale, long Offset, char Imm8 );

	// sar - shift right arithemetic - these shift by the value in register c or by an immediate 8-bit value
	bool SarRegReg16 ( long DestReg );
	bool SarRegReg32 ( long DestReg );
	bool SarRegReg64 ( long DestReg );
	bool SarRegImm16 ( long DestReg, char Imm8 );
	bool SarRegImm32 ( long DestReg, char Imm8 );
	bool SarRegImm64 ( long DestReg, char Imm8 );

	bool SarMemReg32 ( long* DestPtr );
	
	// with memory
	bool SarMemReg16 ( long AddressReg, long IndexReg, long Scale, long Offset );
	bool SarMemReg32 ( long AddressReg, long IndexReg, long Scale, long Offset );
	bool SarMemReg64 ( long AddressReg, long IndexReg, long Scale, long Offset );
	bool SarMemImm16 ( long AddressReg, long IndexReg, long Scale, long Offset, char Imm8 );
	bool SarMemImm32 ( long AddressReg, long IndexReg, long Scale, long Offset, char Imm8 );
	bool SarMemImm64 ( long AddressReg, long IndexReg, long Scale, long Offset, char Imm8 );

	// shr - shift right logical - these shift by the value in register c or by an immediate 8-bit value
	bool ShrRegReg16 ( long DestReg );
	bool ShrRegReg32 ( long DestReg );
	bool ShrRegReg64 ( long DestReg );
	bool ShrRegImm16 ( long DestReg, char Imm8 );
	bool ShrRegImm32 ( long DestReg, char Imm8 );
	bool ShrRegImm64 ( long DestReg, char Imm8 );

	bool ShrMemReg32 ( long* DestPtr );
	
	// ***todo***
	//bool ShrMemImm32 ( long* DestPtr, char Imm8 );
	//bool ShrMemImm32 ( long* DestPtr, char Imm8 );
	
	// with memory
	bool ShrMemReg16 ( long AddressReg, long IndexReg, long Scale, long Offset );
	bool ShrMemReg32 ( long AddressReg, long IndexReg, long Scale, long Offset );
	bool ShrMemReg64 ( long AddressReg, long IndexReg, long Scale, long Offset );
	bool ShrMemImm16 ( long AddressReg, long IndexReg, long Scale, long Offset, char Imm8 );
	bool ShrMemImm32 ( long AddressReg, long IndexReg, long Scale, long Offset, char Imm8 );
	bool ShrMemImm64 ( long AddressReg, long IndexReg, long Scale, long Offset, char Imm8 );

	// sub
	bool SubRegReg16 ( long DestReg, long SrcReg );
	bool SubRegReg32 ( long DestReg, long SrcReg );
	bool SubRegReg64 ( long DestReg, long SrcReg );
	bool SubRegImm16 ( long DestReg, short Imm16 );
	bool SubRegImm32 ( long DestReg, long Imm32 );
	bool SubRegImm64 ( long DestReg, long Imm32 );

	bool SubRegMem16 ( long DestReg, long AddressReg, long IndexReg, long Scale, long Offset );
	bool SubRegMem32 ( long DestReg, long AddressReg, long IndexReg, long Scale, long Offset );
	bool SubRegMem64 ( long DestReg, long AddressReg, long IndexReg, long Scale, long Offset );
	bool SubMemReg16 ( long SrcReg, long AddressReg, long IndexReg, long Scale, long Offset );
	bool SubMemReg32 ( long SrcReg, long AddressReg, long IndexReg, long Scale, long Offset );
	bool SubMemReg64 ( long SrcReg, long AddressReg, long IndexReg, long Scale, long Offset );
	bool SubMemImm16 ( short Imm16, long AddressReg, long IndexReg, long Scale, long Offset );
	bool SubMemImm32 ( long Imm32, long AddressReg, long IndexReg, long Scale, long Offset );
	bool SubMemImm64 ( long Imm32, long AddressReg, long IndexReg, long Scale, long Offset );

	bool SubRegMem16 ( long DestReg, short* SrcPtr );
	bool SubRegMem32 ( long DestReg, long* SrcPtr );
	bool SubRegMem64 ( long DestReg, long long* SrcPtr );
	bool SubMemReg16 ( short* DestPtr, long SrcReg );
	bool SubMemReg32 ( long* DestPtr, long SrcReg );
	bool SubMemReg64 ( long long* DestPtr, long SrcReg );
	bool SubMemImm16 ( short* DestPtr, short Imm16 );
	bool SubMemImm32 ( long* DestPtr, long Imm32 );
	bool SubMemImm64 ( long long* DestPtr, long Imm32 );

	// 8-bit immediate
	bool SubReg16Imm8 ( long DestReg, char Imm8 );
	bool SubReg32Imm8 ( long DestReg, char Imm8 );
	bool SubReg64Imm8 ( long DestReg, char Imm8 );

	bool SubMem16Imm8 ( short* DestPtr, char Imm8 );
	bool SubMem32Imm8 ( long* DestPtr, char Imm8 );
	bool SubMem64Imm8 ( long long* DestPtr, char Imm8 );
	
	// add with accumulator
	bool SubAcc16Imm16 ( short Imm16 );
	bool SubAcc32Imm32 ( long Imm32 );
	bool SubAcc64Imm32 ( long Imm32 );
	
	// self-optimizing versions
	bool SubReg16ImmX ( long DestReg, short ImmX );
	bool SubReg32ImmX ( long DestReg, long ImmX );
	bool SubReg64ImmX ( long DestReg, long ImmX );
	
	// self-optimizing
	bool SubMem16ImmX ( short* DestPtr, short ImmX );
	bool SubMem32ImmX ( long* DestPtr, long ImmX );
	bool SubMem64ImmX ( long long* DestPtr, long ImmX );

	// sbb - subtract with borrow
	bool SbbRegReg16(long DestReg, long SrcReg);
	bool SbbRegReg32(long DestReg, long SrcReg);
	bool SbbRegReg64(long DestReg, long SrcReg);

	bool SbbReg16Imm8(long DestReg, char Imm8);
	bool SbbReg32Imm8(long DestReg, char Imm8);
	bool SbbReg64Imm8(long DestReg, char Imm8);

	
	// test - perform and without writing result and just setting the flags

	bool TestRegReg16 ( long DestReg, long SrcReg );
	bool TestRegReg32 ( long DestReg, long SrcReg );
	bool TestRegReg64 ( long DestReg, long SrcReg );
	
	bool TestRegImm16 ( long DestReg, short Imm16 );
	bool TestRegImm32 ( long DestReg, long Imm32 );
	bool TestRegImm64 ( long DestReg, long Imm32 );
	
	bool TestMemReg16 ( short* DestPtr, long SrcReg );
	bool TestMemReg32 ( long* DestPtr, long SrcReg );
	bool TestMemReg64 ( long long* DestPtr, long SrcReg );

	bool TestMemImm16 ( short* DestPtr, short Imm16 );
	bool TestMemImm32 ( long* DestPtr, long Imm32 );
	bool TestMemImm64 ( long long* DestPtr, long Imm32 );

	bool TestRegMem16 ( long DestReg, long AddressReg, long IndexReg, long Scale, long Offset );
	bool TestRegMem32 ( long DestReg, long AddressReg, long IndexReg, long Scale, long Offset );
	bool TestRegMem64 ( long DestReg, long AddressReg, long IndexReg, long Scale, long Offset );

	
	// 8-bit immediate
	bool TestReg16Imm8 ( long DestReg, char Imm8 );
	bool TestReg32Imm8 ( long DestReg, char Imm8 );
	bool TestReg64Imm8 ( long DestReg, char Imm8 );

	bool TestMem16Imm8 ( short* DestPtr, char Imm8 );
	bool TestMem32Imm8 ( long* DestPtr, char Imm8 );
	bool TestMem64Imm8 ( long long* DestPtr, char Imm8 );
	
	// test with accumulator
	bool TestAcc16Imm16 ( short Imm16 );
	bool TestAcc32Imm32 ( long Imm32 );
	bool TestAcc64Imm32 ( long Imm32 );
	
	// self-optimizing versions
	bool TestReg16ImmX ( long DestReg, short ImmX );
	bool TestReg32ImmX ( long DestReg, long ImmX );
	bool TestReg64ImmX ( long DestReg, long ImmX );
	
	// self-optimizing
	bool TestMem16ImmX ( short* DestPtr, short ImmX );
	bool TestMem32ImmX ( long* DestPtr, long ImmX );
	bool TestMem64ImmX ( long long* DestPtr, long ImmX );
	
	
	// xchg - exchange register with register or memory
	bool XchgRegReg16 ( long DestReg, long SrcReg );
	bool XchgRegReg32 ( long DestReg, long SrcReg );
	bool XchgRegReg64 ( long DestReg, long SrcReg );

	bool XchgRegMem16 ( long DestReg, long AddressReg, long IndexReg, long Scale, long Offset );
	bool XchgRegMem32 ( long DestReg, long AddressReg, long IndexReg, long Scale, long Offset );
	bool XchgRegMem64 ( long DestReg, long AddressReg, long IndexReg, long Scale, long Offset );

	// xor
	bool XorRegReg16 ( long DestReg, long SrcReg );
	bool XorRegReg32 ( long DestReg, long SrcReg );
	bool XorRegReg64 ( long DestReg, long SrcReg );
	bool XorRegImm16 ( long DestReg, short Imm16 );
	bool XorRegImm32 ( long DestReg, long Imm32 );
	bool XorRegImm64 ( long DestReg, long Imm32 );

	bool XorRegMem8 ( long DestReg, long AddressReg, long IndexReg, long Scale, long Offset );
	bool XorRegMem16 ( long DestReg, long AddressReg, long IndexReg, long Scale, long Offset );
	bool XorRegMem32 ( long DestReg, long AddressReg, long IndexReg, long Scale, long Offset );
	bool XorRegMem64 ( long DestReg, long AddressReg, long IndexReg, long Scale, long Offset );
	bool XorMemReg16 ( long SrcReg, long AddressReg, long IndexReg, long Scale, long Offset );
	bool XorMemReg32 ( long SrcReg, long AddressReg, long IndexReg, long Scale, long Offset );
	bool XorMemReg64 ( long SrcReg, long AddressReg, long IndexReg, long Scale, long Offset );
	bool XorMemImm8 ( char Imm8, long AddressReg, long IndexReg, long Scale, long Offset );
	bool XorMemImm16 ( short Imm16, long AddressReg, long IndexReg, long Scale, long Offset );
	bool XorMemImm32 ( long Imm32, long AddressReg, long IndexReg, long Scale, long Offset );
	bool XorMemImm64 ( long Imm32, long AddressReg, long IndexReg, long Scale, long Offset );

	bool XorRegMem16 ( long DestReg, short* SrcPtr );
	bool XorRegMem32 ( long DestReg, long* SrcPtr );
	bool XorRegMem64 ( long DestReg, long long* SrcPtr );
	bool XorMemReg16 ( short* DestPtr, long SrcReg );
	bool XorMemReg32 ( long* DestPtr, long SrcReg );
	bool XorMemReg64 ( long long* DestPtr, long SrcReg );
	bool XorMemImm16 ( short* DestPtr, short Imm16 );
	bool XorMemImm32 ( long* DestPtr, long Imm32 );
	bool XorMemImm64 ( long long* DestPtr, long Imm32 );

	// 8-bit immediate
	bool XorReg16Imm8 ( long DestReg, char Imm8 );
	bool XorReg32Imm8 ( long DestReg, char Imm8 );
	bool XorReg64Imm8 ( long DestReg, char Imm8 );

	bool XorMem16Imm8 ( short* DestPtr, char Imm8 );
	bool XorMem32Imm8 ( long* DestPtr, char Imm8 );
	bool XorMem64Imm8 ( long long* DestPtr, char Imm8 );
	
	// add with accumulator
	bool XorAcc16Imm16 ( short Imm16 );
	bool XorAcc32Imm32 ( long Imm32 );
	bool XorAcc64Imm32 ( long Imm32 );
	
	// self-optimizing versions
	bool XorReg16ImmX ( long DestReg, short ImmX );
	bool XorReg32ImmX ( long DestReg, long ImmX );
	bool XorReg64ImmX ( long DestReg, long ImmX );
	
	// self-optimizing
	bool XorMem16ImmX ( short* DestPtr, short ImmX );
	bool XorMem32ImmX ( long* DestPtr, long ImmX );
	bool XorMem64ImmX ( long long* DestPtr, long ImmX );
	
	
	// ** SSE/AVX Register Instructions ** //


	// mxcsr
	bool stmxcsr(long* pDstPtr);
	bool vstmxcsr(long* pDstPtr);

	bool ldmxcsr(long* pDstPtr);
	bool vldmxcsr(long* pDstPtr);

	// movhlps - combine high quadwords of xmm registers, 3rd operand on bottom

	// avx
	bool movhlps2 ( long sseDestReg, long sseSrc1Reg, long sseSrc2Reg );
	
	// movlhps - combine low quadwords of xmm registers, 3rd operand on top

	// avx
	bool movlhps2 ( long sseDestReg, long sseSrc1Reg, long sseSrc2Reg );

	// pinsr

	// avx
	bool pinsrb1regreg ( long sseDestReg, long sseSrcReg1, long x64SrcReg2, char Imm8 );
	bool pinsrw1regreg( long sseDestReg, long sseSrcReg1, long x64SrcReg2, char Imm8 );
	bool pinsrd1regreg( long sseDestReg, long sseSrcReg1, long x64SrcReg2, char Imm8 );
	bool pinsrq1regreg( long sseDestReg, long sseSrcReg1, long x64SrcReg2, char Imm8 );

	bool pinsrb1regmem(long sseDestReg, long sseSrcReg1, char* pSrcPtr, char Imm8);
	bool pinsrw1regmem(long sseDestReg, long sseSrcReg1, short* pSrcPtr, char Imm8);
	bool pinsrd1regmem(long sseDestReg, long sseSrcReg1, long* pSrcPtr, char Imm8);
	bool pinsrq1regmem(long sseDestReg, long sseSrcReg1, long long* pSrcPtr, char Imm8);

	bool vinserti128regreg(long sseDestReg, long sseSrcReg1, long sseSrcReg2, char Imm8);
	bool vinserti128regmem(long sseDestReg, long sseSrcReg1, void* SrcPtr, char Imm8);

	// avx512
	bool vinserti32x4_rri256(long avxDestReg, long avxSrcReg1, long sseSrcReg2, char Imm8, long mask = 0, long z = 0);
	bool vinserti32x4_rri512(long avxDestReg, long avxSrcReg1, long sseSrcReg2, char Imm8, long mask = 0, long z = 0);

	bool vinserti32x8_rri512(long avxDestReg, long avxSrcReg1, long sseSrcReg2, char Imm8, long mask = 0, long z = 0);

	bool vinserti32x4_rmi256(long dst, long src0, void* SrcPtr, char Imm8, long mask = 0, long z = 0);
	bool vinserti32x4_rmi512(long dst, long src0, void* SrcPtr, char Imm8, long mask = 0, long z = 0);

	bool vinsertb_rri128(long dst, long src0, long x64reg, char Imm8);
	bool vinsertd_rri128(long dst, long src0, long x64reg, char Imm8);
	bool vinsertq_rri128(long dst, long src0, long x64reg, char Imm8);

	bool vinsertb_rmi128(long dst, long src0, void* SrcPtr, char Imm8);
	bool vinsertd_rmi128(long dst, long src0, void* SrcPtr, char Imm8);
	bool vinsertq_rmi128(long dst, long src0, void* SrcPtr, char Imm8);

	// pextr

	// avx
	bool pextrb1regreg ( long x64DestReg, long sseSrcReg, char Imm8 );
	bool pextrw1regreg( long x64DestReg, long sseSrcReg, char Imm8 );
	bool pextrd1regreg( long x64DestReg, long sseSrcReg, char Imm8 );
	bool pextrq1regreg( long x64DestReg, long sseSrcReg, char Imm8 );

	bool pextrd1memreg(long* pDstPtr, long sseSrcReg, char Imm8);
	bool pextrq1memreg(long long* pDstPtr, long sseSrcReg, char Imm8);

	bool vextracti128regreg(long sseDestReg, long sseSrcReg, char Imm8);

	bool vextracti128memreg(void* DstPtr, long sseSrcReg, char Imm8);

	// avx512
	bool vextracti32x4_rri256(long dst, long src0, char Imm8, long mask = 0, long z = 0);
	bool vextracti32x4_rri512(long dst, long src0, char Imm8, long mask = 0, long z = 0);

	bool vextracti64x4_rri512(long dst, long src0, char Imm8, long mask = 0, long z = 0);

	bool vextracti32x4_mri256(void* DstPtr, long src0, char Imm8, long mask = 0, long z = 0);
	bool vextracti32x4_mri512(void* DstPtr, long src0, char Imm8, long mask = 0, long z = 0);

	bool vextractb_rri128(long x64reg, long src0, char Imm8);
	bool vextractd_rri128(long x64reg, long src0, char Imm8);
	bool vextractq_rri128(long x64reg, long src0, char Imm8);

	bool vextractb_mri128(void* DstPtr, long src0, char Imm8);
	bool vextractd_mri128(void* DstPtr, long src0, char Imm8);
	bool vextractq_mri128(void* DstPtr, long src0, char Imm8);

	// blendvps
	//bool blendvps ( long sseDestReg, long sseSrcReg1, long x64SrcReg2, long x64SrcReg3 );

	// vpblendmd - blend 32-bit double-words using mask

	// avx512
	
	bool vpblendmd_rrr512(long dst, long src0, long src1, long mask = 0, long z = 0);
	bool vpblendmd_rrr256(long dst, long src0, long src1, long mask = 0, long z = 0);
	bool vpblendmd_rrr128(long dst, long src0, long src1, long mask = 0, long z = 0);

	// vpblendmq - blend 64-bit quad-words using mask

	// avx512

	bool vpblendmq_rrr512(long dst, long src0, long src1, long mask = 0, long z = 0);
	bool vpblendmq_rrr256(long dst, long src0, long src1, long mask = 0, long z = 0);
	bool vpblendmq_rrr128(long dst, long src0, long src1, long mask = 0, long z = 0);



	// pblendw

	// avx
	bool pblendw1regregimm(long sseDestReg, long sseSrcReg1, long sseSrcReg2, char Imm8);
	bool pblendw2regregimm (long sseDestReg, long sseSrcReg1, long sseSrcReg2, char Imm8);
	
	bool pblendw1regmemimm(long sseDestReg, long sseSrcReg1, void* SrcPtr, char Imm8);

	// sse
	bool pblendwregregimm ( long sseDestReg, long sseSrcReg, char Imm8 );
	bool pblendwregmemimm ( long sseDestReg, void* SrcPtr, char Imm8 );
	
	// pblendvb

	// avx
	bool pblendvb1regreg(long sseDestReg, long sseSrcReg1, long sseSrcReg2, long sseSrcReg3);
	bool pblendvb2regreg(long sseDestReg, long sseSrcReg1, long sseSrcReg2, long sseSrcReg3);
	//bool pblendvb1regmem(long sseDestReg, void* SrcPtr);


	// sse
	bool pblendvbregreg ( long sseDestReg, long sseSrcReg );
	bool pblendvbregmem ( long sseDestReg, void* SrcPtr );


	// knot
	bool knotb(long dst, long src0);
	bool knotw(long dst, long src0);
	bool knotd(long dst, long src0);
	bool knotq(long dst, long src0);

	// kxnor
	bool kxnorb(long dst, long src0, long src1);
	bool kxnorw(long dst, long src0, long src1);
	bool kxnord(long dst, long src0, long src1);
	bool kxnorq(long dst, long src0, long src1);

	// kxor
	bool korb(long dst, long src0, long src1);
	bool korw(long dst, long src0, long src1);
	bool kord(long dst, long src0, long src1);
	bool korq(long dst, long src0, long src1);

	// kand
	bool kandb(long dst, long src0, long src1);
	bool kandw(long dst, long src0, long src1);
	bool kandd(long dst, long src0, long src1);
	bool kandq(long dst, long src0, long src1);

	// kandn
	bool kandnb(long dst, long src0, long src1);
	bool kandnw(long dst, long src0, long src1);
	bool kandnd(long dst, long src0, long src1);
	bool kandnq(long dst, long src0, long src1);

	// kmov
	bool kmovbmskreg(long dstmsk, long srcreg0);
	bool kmovbregmsk(long dstreg, long srcmsk0);
	bool kmovbmskmem(long dstmsk, char* srcptr0);

	bool kmovwmskreg(long dstmsk, long srcreg0);
	bool kmovwregmsk(long dstreg, long srcmsk0);
	bool kmovwmskmem(long dstmsk, short* srcptr0);

	bool kmovdmskreg(long dstmsk, long srcreg0);
	bool kmovdregmsk(long dstreg, long srcmsk0);
	bool kmovdmskmem(long dstmsk, long* srcptr0);
	bool kmovdmemmsk(long* dstptr0, long srcmsk0);

	bool kmovqmskreg(long dstmsk, long srcreg0);
	bool kmovqregmsk(long dstreg, long srcmsk0);
	bool kmovqmskmem(long dstmsk, long long* srcptr0);


	// kshiftr - shift right k register
	bool kshiftrb(long dst, long src0, char Imm8);
	bool kshiftrw(long dst, long src0, char Imm8);
	bool kshiftrd(long dst, long src0, char Imm8);
	bool kshiftrq(long dst, long src0, char Imm8);

	// kshiftl - shift right k register
	bool kshiftlb(long dst, long src0, char Imm8);
	bool kshiftlw(long dst, long src0, char Imm8);
	bool kshiftld(long dst, long src0, char Imm8);
	bool kshiftlq(long dst, long src0, char Imm8);

	// kunpack - src1 goes on top, src2 on the bottom
	bool kunpackbw(long dst, long src0, long src1);
	bool kunpackwd(long dst, long src0, long src1);
	bool kunpackdq(long dst, long src0, long src1);

	// movaps
	bool movaps128 ( long sseDestReg, long sseSrcReg );
	bool movaps_to_mem128 ( void* DestPtr, long sseSrcReg );
	bool movaps_from_mem128 ( long sseDestReg, void* SrcPtr );
	bool movaps_to_mem128 ( long sseSrcReg, long x64BaseReg, long x64IndexReg, long Scale, long Offset );
	bool movaps_from_mem128 ( long sseDestReg, long x64BaseReg, long x64IndexReg, long Scale, long Offset );
	

	// movdqa

	// avx 128-bit
	bool movdqa1_regreg ( long sseDestReg, long sseSrcReg );
	bool movdqa1_regmem ( long sseDestReg, void* SrcPtr );
	bool movdqa1_memreg ( void* DestPtr, long sseSrcReg );
	bool movdqa1_memreg ( long sseSrcReg, long x64BaseReg, long x64IndexReg, long Scale, long Offset );
	bool movdqa1_regmem ( long sseDestReg, long x64BaseReg, long x64IndexReg, long Scale, long Offset );

	bool movdqu1_memreg(void* DestPtr, long sseSrcReg);
	bool movdqu1_memreg(long sseSrcReg, long x64BaseReg, long x64IndexReg, long Scale, long Offset);
	bool movdqu1_regmem(long sseDestReg, long x64BaseReg, long x64IndexReg, long Scale, long Offset);

	// avx 256-bit
	bool movdqa2_regreg ( long sseDestReg, long sseSrcReg );
	bool movdqa2_regmem ( long sseDestReg, void* SrcPtr );
	bool movdqa2_memreg ( void* DestPtr, long sseSrcReg );
	bool movdqa2_memreg ( long sseSrcReg, long x64BaseReg, long x64IndexReg, long Scale, long Offset );
	bool movdqa2_regmem ( long sseDestReg, long x64BaseReg, long x64IndexReg, long Scale, long Offset );
	
	// movsd - move single 32-bit value from memory or to memory
	bool movsd1_regmem(long sseDstReg, long long* pSrcPtr);
	bool movsd1_memreg(long long* pDstPtr, long sseSrcReg);


	//bool movdqa_to_mem128 ( long sseSrcReg, long x64BaseReg, long x64IndexReg, long Scale, long Offset );
	//bool movdqa_from_mem128 ( long sseDestReg, long x64BaseReg, long x64IndexReg, long Scale, long Offset );
	//bool movdqa256 ( long sseDestReg, long sseSrcReg );
	//bool movdqa_to_mem256 ( long sseSrcReg, long x64BaseReg, long x64IndexReg, long Scale, long Offset );
	//bool movdqa_from_mem256 ( long sseDestReg, long x64BaseReg, long x64IndexReg, long Scale, long Offset );

	// sse
	bool movdqa_regreg ( long sseDestReg, long sseSrcReg );
	bool movdqa_memreg ( void* DestPtr, long sseSrcReg );
	bool movdqa_regmem ( long sseDestReg, void* SrcPtr );
	bool movdqa_to_mem128 ( long sseSrcReg, long x64BaseReg, long x64IndexReg, long Scale, long Offset );
	bool movdqa_from_mem128 ( long sseDestReg, long x64BaseReg, long x64IndexReg, long Scale, long Offset );
	
	bool movdqu_regreg ( long sseDestReg, long sseSrcReg );
	bool movdqu_memreg ( void* DestPtr, long sseSrcReg );
	bool movdqu_regmem ( long sseDestReg, void* SrcPtr );
	bool movdqu_to_mem128 ( long sseSrcReg, long x64BaseReg, long x64IndexReg, long Scale, long Offset );
	bool movdqu_from_mem128 ( long sseDestReg, long x64BaseReg, long x64IndexReg, long Scale, long Offset );
	
	// move 32-bit value into lower part of xmm register (zero upper bits)
	//bool movd_regmem ( long sseDestReg, long* SrcPtr );

	// move 64-bit value into lower part of xmm register (zero upper bits)
	//bool movq_regmem ( long sseDestReg, long long* SrcPtr );
	
	// movss - move single 32-bit value from memory or to memory
	bool movss_to_mem128 ( long sseSrcReg, long x64BaseReg, long x64IndexReg, long Scale, long Offset );
	bool movss_from_mem128 ( long sseDestReg, long x64BaseReg, long x64IndexReg, long Scale, long Offset );


	// avx512
	bool movdqa32_rr512(long dst, long src0, long mask = 0, long z = 0);
	bool movdqa32_rr256(long dst, long src0, long mask = 0, long z = 0);
	bool movdqa32_rr128(long dst, long src0, long mask = 0, long z = 0);

	bool movdqa64_rr512(long dst, long src0, long mask = 0, long z = 0);
	bool movdqa64_rr256(long dst, long src0, long mask = 0, long z = 0);
	bool movdqa64_rr128(long dst, long src0, long mask = 0, long z = 0);

	bool movdqa32_rm512(long dst, void* pSrcPtr, long mask = 0, long z = 0);
	bool movdqa32_rm256(long dst, void* pSrcPtr, long mask = 0, long z = 0);
	bool movdqa32_rm128(long dst, void* pSrcPtr, long mask = 0, long z = 0);

	bool movdqa32_mr512(void* pDstPtr, long dst, long mask = 0, long z = 0);
	bool movdqa32_mr256(void* pDstPtr, long dst, long mask = 0, long z = 0);
	bool movdqa32_mr128(void* pDstPtr, long dst, long mask = 0, long z = 0);

	// pabs

	// avx512
	bool pabsd_rr512(long dst, long src0, long mask = 0, long z = 0);
	bool pabsd_rr256(long dst, long src0, long mask = 0, long z = 0);
	bool pabsd_rr128(long dst, long src0, long mask = 0, long z = 0);

	// avx 128-bit
	bool pabsb1regreg ( long sseDestReg, long sseSrcReg );
	bool pabsw1regreg ( long sseDestReg, long sseSrcReg );
	bool pabsd1regreg ( long sseDestReg, long sseSrcReg );
	
	bool pabsb1regmem ( long sseDestReg, void* SrcPtr );
	bool pabsw1regmem ( long sseDestReg, void* SrcPtr );
	bool pabsd1regmem ( long sseDestReg, void* SrcPtr );

	// avx 256-bit
	bool pabsb2regreg(long sseDestReg, long sseSrcReg);
	bool pabsw2regreg(long sseDestReg, long sseSrcReg);
	bool pabsd2regreg(long sseDestReg, long sseSrcReg);

	bool pabsb2regmem(long sseDestReg, void* SrcPtr);
	bool pabsw2regmem(long sseDestReg, void* SrcPtr);
	bool pabsd2regmem(long sseDestReg, void* SrcPtr);

	// sse
	bool pabswregreg ( long sseDestReg, long sseSrcReg );
	bool pabswregmem ( long sseDestReg, void* SrcPtr );
	bool pabsdregreg ( long sseDestReg, long sseSrcReg );
	bool pabsdregmem ( long sseDestReg, void* SrcPtr );

	// packs

	// avx512

	// rrr - register-register-register
	bool packsdw_rrr512(long dst, long src0, long src1, long mask = 0, long z = 0);
	bool packsdw_rrr256(long dst, long src0, long src1, long mask = 0, long z = 0);
	bool packsdw_rrr128(long dst, long src0, long src1, long mask = 0, long z = 0);

	bool packswb_rrr512(long dst, long src0, long src1, long mask = 0, long z = 0);
	bool packswb_rrr256(long dst, long src0, long src1, long mask = 0, long z = 0);
	bool packswb_rrr128(long dst, long src0, long src1, long mask = 0, long z = 0);


	// sse
	bool packsdwregreg(long sseDestReg, long sseSrcReg);
	bool packswbregreg(long sseDestReg, long sseSrcReg);

	// avx128
	bool packsdw1regreg(long sseDestReg, long sseSrc1Reg, long sseSrc2Reg);
	bool packswb1regreg(long sseDestReg, long sseSrc1Reg, long sseSrc2Reg);

	// avx2 - 256
	bool packsdw2regreg(long sseDestReg, long sseSrc1Reg, long sseSrc2Reg);
	bool packswb2regreg(long sseDestReg, long sseSrc1Reg, long sseSrc2Reg);

	// packus

	// avx512

	// rrr - register-register-register
	bool packusdw_rrr512(long dst, long src0, long src1, long mask = 0, long z = 0);
	bool packusdw_rrr256(long dst, long src0, long src1, long mask = 0, long z = 0);
	bool packusdw_rrr128(long dst, long src0, long src1, long mask = 0, long z = 0);

	bool packuswb_rrr512(long dst, long src0, long src1, long mask = 0, long z = 0);
	bool packuswb_rrr256(long dst, long src0, long src1, long mask = 0, long z = 0);
	bool packuswb_rrr128(long dst, long src0, long src1, long mask = 0, long z = 0);

	// avx128
	bool packusdw1regreg ( long sseDestReg, long sseSrc1Reg, long sseSrc2Reg );
	bool packuswb1regreg ( long sseDestReg, long sseSrc1Reg, long sseSrc2Reg );

	bool packsswb1regreg(long sseDestReg, long sseSrc1Reg, long sseSrcReg);
	bool packsswb1regmem(long sseDestReg, long sseSrc1Reg, void* SrcPtr);
	bool packssdw1regreg(long sseDestReg, long sseSrc1Reg, long sseSrcReg);
	bool packssdw1regmem(long sseDestReg, long sseSrc1Reg, void* SrcPtr);

	// avx2 - 256
	bool packusdw2regreg(long sseDestReg, long sseSrc1Reg, long sseSrc2Reg);
	bool packuswb2regreg(long sseDestReg, long sseSrc1Reg, long sseSrc2Reg);

	// sse
	bool packuswbregreg ( long sseDestReg, long sseSrcReg );
	bool packuswbregmem ( long sseDestReg, void* SrcPtr );
	bool packusdwregreg ( long sseDestReg, long sseSrcReg );
	bool packusdwregmem ( long sseDestReg, void* SrcPtr );
	
	bool packsswbregreg ( long sseDestReg, long sseSrcReg );
	bool packsswbregmem ( long sseDestReg, void* SrcPtr );
	bool packssdwregreg ( long sseDestReg, long sseSrcReg );
	bool packssdwregmem ( long sseDestReg, void* SrcPtr );
	
	// pmovzx

	// avx
	bool pmovzxbw1regreg ( long sseDestReg, long sseSrcReg );
	bool pmovzxbw1regmem ( long sseDestReg, void* SrcPtr );
	bool pmovzxbd1regreg ( long sseDestReg, long sseSrcReg );
	bool pmovzxbd1regmem ( long sseDestReg, void* SrcPtr );
	bool pmovzxbq1regreg ( long sseDestReg, long sseSrcReg );
	bool pmovzxbq1regmem ( long sseDestReg, void* SrcPtr );
	bool pmovzxwd1regreg ( long sseDestReg, long sseSrcReg );
	bool pmovzxwd1regmem ( long sseDestReg, void* SrcPtr );
	bool pmovzxwq1regreg ( long sseDestReg, long sseSrcReg );
	bool pmovzxwq1regmem ( long sseDestReg, void* SrcPtr );
	bool pmovzxdq1regreg ( long sseDestReg, long sseSrcReg );
	bool pmovzxdq1regmem ( long sseDestReg, void* SrcPtr );

	bool pmovzxbw2regreg ( long sseDestReg, long sseSrcReg );
	bool pmovzxbw2regmem ( long sseDestReg, void* SrcPtr );
	bool pmovzxbd2regreg ( long sseDestReg, long sseSrcReg );
	bool pmovzxbd2regmem ( long sseDestReg, void* SrcPtr );
	bool pmovzxbq2regreg ( long sseDestReg, long sseSrcReg );
	bool pmovzxbq2regmem ( long sseDestReg, void* SrcPtr );
	bool pmovzxwd2regreg ( long sseDestReg, long sseSrcReg );
	bool pmovzxwd2regmem ( long sseDestReg, void* SrcPtr );
	bool pmovzxwq2regreg ( long sseDestReg, long sseSrcReg );
	bool pmovzxwq2regmem ( long sseDestReg, void* SrcPtr );
	bool pmovzxdq2regreg ( long sseDestReg, long sseSrcReg );
	bool pmovzxdq2regmem ( long sseDestReg, void* SrcPtr );
	
	
	
	// sse
	bool pmovzxbwregreg ( long sseDestReg, long sseSrcReg );
	bool pmovzxbwregmem ( long sseDestReg, void* SrcPtr );
	bool pmovzxbdregreg ( long sseDestReg, long sseSrcReg );
	bool pmovzxbdregmem ( long sseDestReg, void* SrcPtr );
	bool pmovzxbqregreg ( long sseDestReg, long sseSrcReg );
	bool pmovzxbqregmem ( long sseDestReg, void* SrcPtr );
	bool pmovzxwdregreg ( long sseDestReg, long sseSrcReg );
	bool pmovzxwdregmem ( long sseDestReg, void* SrcPtr );
	bool pmovzxwqregreg ( long sseDestReg, long sseSrcReg );
	bool pmovzxwqregmem ( long sseDestReg, void* SrcPtr );
	bool pmovzxdqregreg ( long sseDestReg, long sseSrcReg );
	bool pmovzxdqregmem ( long sseDestReg, void* SrcPtr );
	

	// padd

	// avx512

	// rrm - register-register-memory
	bool paddd_rrm512(long dst, long src0, void* pSrcPtr, long mask = 0, long z = 0);
	bool paddd_rrm256(long dst, long src0, void* pSrcPtr, long mask = 0, long z = 0);
	bool paddd_rrm128(long dst, long src0, void* pSrcPtr, long mask = 0, long z = 0);

	bool paddq_rrm512(long dst, long src0, void* pSrcPtr, long mask = 0, long z = 0);
	bool paddq_rrm256(long dst, long src0, void* pSrcPtr, long mask = 0, long z = 0);
	bool paddq_rrm128(long dst, long src0, void* pSrcPtr, long mask = 0, long z = 0);

	// rrb - register-register-broadcast memory
	bool paddd_rrb512(long dst, long src0, long* pSrcPtr, long mask = 0, long z = 0);
	bool paddd_rrb256(long dst, long src0, long* pSrcPtr, long mask = 0, long z = 0);
	bool paddd_rrb128(long dst, long src0, long* pSrcPtr, long mask = 0, long z = 0);

	bool paddq_rrb512(long dst, long src0, long long* pSrcPtr, long mask = 0, long z = 0);
	bool paddq_rrb256(long dst, long src0, long long* pSrcPtr, long mask = 0, long z = 0);
	bool paddq_rrb128(long dst, long src0, long long* pSrcPtr, long mask = 0, long z = 0);

	// rrr - register-register-register
	bool paddd_rrr512(long dst, long src0, long src1, long mask = 0, long z = 0);
	bool paddd_rrr256(long dst, long src0, long src1, long mask = 0, long z = 0);
	bool paddd_rrr128(long dst, long src0, long src1, long mask = 0, long z = 0);

	bool paddq_rrr512(long dst, long src0, long src1, long mask = 0, long z = 0);
	bool paddq_rrr256(long dst, long src0, long src1, long mask = 0, long z = 0);
	bool paddq_rrr128(long dst, long src0, long src1, long mask = 0, long z = 0);


	// avx 128-bit
	bool paddb1regreg ( long sseDestReg, long sseSrc1Reg, long sseSrc2Reg );
	bool paddw1regreg ( long sseDestReg, long sseSrc1Reg, long sseSrc2Reg );
	bool paddd1regreg ( long sseDestReg, long sseSrc1Reg, long sseSrc2Reg );
	bool paddq1regreg(long sseDestReg, long sseSrc1Reg, long sseSrc2Reg);

	bool paddb1regmem(long sseDestReg, long sseSrc1Reg, void* pSrcPtr);
	bool paddw1regmem(long sseDestReg, long sseSrc1Reg, void* pSrcPtr);
	bool paddd1regmem(long sseDestReg, long sseSrc1Reg, void* pSrcPtr);
	bool paddq1regmem(long sseDestReg, long sseSrc1Reg, void* pSrcPtr);

	// avx 256-bit
	bool paddb2regreg(long sseDestReg, long sseSrc1Reg, long sseSrc2Reg);
	bool paddw2regreg(long sseDestReg, long sseSrc1Reg, long sseSrc2Reg);
	bool paddd2regreg(long sseDestReg, long sseSrc1Reg, long sseSrc2Reg);
	bool paddq2regreg(long sseDestReg, long sseSrc1Reg, long sseSrc2Reg);

	// sse
	bool paddbregreg ( long sseDestReg, long sseSrcReg );
	bool paddwregreg ( long sseDestReg, long sseSrcReg );
	bool padddregreg ( long sseDestReg, long sseSrcReg );
	bool paddqregreg ( long sseDestReg, long sseSrcReg );

	bool paddbregmem ( long sseDestReg, void* SrcPtr );
	bool paddwregmem ( long sseDestReg, void* SrcPtr );
	bool padddregmem ( long sseDestReg, void* SrcPtr );
	bool paddqregmem ( long sseDestReg, void* SrcPtr );

	// paddus

	// avx 128-bit
	bool paddusb1regreg ( long sseDestReg, long sseSrc1Reg, long sseSrc2Reg );
	bool paddusw1regreg ( long sseDestReg, long sseSrc1Reg, long sseSrc2Reg );

	bool paddusb1regmem(long sseDestReg, long sseSrc1Reg, void* pSrcPtr);
	bool paddusw1regmem(long sseDestReg, long sseSrc1Reg, void* pSrcPtr);

	// sse
	bool paddusbregreg ( long sseDestReg, long sseSrcReg );
	bool paddusbregmem ( long sseDestReg, void* SrcPtr );
	bool padduswregreg ( long sseDestReg, long sseSrcReg );
	bool padduswregmem ( long sseDestReg, void* SrcPtr );
	
	// padds - add packed signed values with saturation

	// avx 128-bit
	bool paddsb1regreg ( long sseDestReg, long sseSrc1Reg, long sseSrc2Reg );
	bool paddsw1regreg ( long sseDestReg, long sseSrc1Reg, long sseSrc2Reg );

	bool paddsb1regmem(long sseDestReg, long sseSrc1Reg, void* pSrcPtr);
	bool paddsw1regmem(long sseDestReg, long sseSrc1Reg, void* pSrcPtr);

	// sse
	bool paddsbregreg ( long sseDestReg, long sseSrcReg );
	bool paddsbregmem ( long sseDestReg, void* SrcPtr );
	bool paddswregreg ( long sseDestReg, long sseSrcReg );
	bool paddswregmem ( long sseDestReg, void* SrcPtr );
	
	// pand

	// avx512

	// rrr - register-register-register
	bool pand_rrr512(long dst, long src0, long src1, long mask = 0, long z = 0);
	bool pand_rrr256(long dst, long src0, long src1, long mask = 0, long z = 0);
	bool pand_rrr128(long dst, long src0, long src1, long mask = 0, long z = 0);

	// rrm - register-register-memory
	bool pandd_rrm512(long dst, long src0, void* pSrcPtr, long mask = 0, long z = 0);
	bool pandd_rrm256(long dst, long src0, void* pSrcPtr, long mask = 0, long z = 0);
	bool pandd_rrm128(long dst, long src0, void* pSrcPtr, long mask = 0, long z = 0);

	bool pandq_rrm512(long dst, long src0, void* pSrcPtr, long mask = 0, long z = 0);
	bool pandq_rrm256(long dst, long src0, void* pSrcPtr, long mask = 0, long z = 0);
	bool pandq_rrm128(long dst, long src0, void* pSrcPtr, long mask = 0, long z = 0);

	// rrb - register-register-broadcast memory
	bool pandd_rrb512(long dst, long src0, long* pSrcPtr, long mask = 0, long z = 0);
	bool pandd_rrb256(long dst, long src0, long* pSrcPtr, long mask = 0, long z = 0);
	bool pandd_rrb128(long dst, long src0, long* pSrcPtr, long mask = 0, long z = 0);

	bool pandq_rrb512(long dst, long src0, long long* pSrcPtr, long mask = 0, long z = 0);
	bool pandq_rrb256(long dst, long src0, long long* pSrcPtr, long mask = 0, long z = 0);
	bool pandq_rrb128(long dst, long src0, long long* pSrcPtr, long mask = 0, long z = 0);


	// avx 128-bit
	bool pand1regreg ( long sseDestReg, long sseSrc1Reg, long sseSrc2Reg );

	bool pand1regmem(long sseDestReg, long sseSrc1Reg, void* pSrcPtr);

	// avx 256-bit
	bool pand2regreg(long sseDestReg, long sseSrc1Reg, long sseSrc2Reg);

	// sse
	bool pandregreg ( long sseDestReg, long sseSrcReg );
	bool pandregmem ( long sseDestReg, void* SrcPtr );
	
	// pandn

	// avx512

	// rrr - register-register-register
	bool pandn_rrr512(long dst, long src0, long src1, long mask = 0, long z = 0);
	bool pandn_rrr256(long dst, long src0, long src1, long mask = 0, long z = 0);
	bool pandn_rrr128(long dst, long src0, long src1, long mask = 0, long z = 0);

	// avx 128-bit
	bool pandn1regreg ( long sseDestReg, long sseSrc1Reg, long sseSrc2Reg );

	// avx 256-bit
	bool pandn2regreg(long sseDestReg, long sseSrc1Reg, long sseSrc2Reg);

	// sse
	bool pandnregreg ( long sseDestReg, long sseSrcReg );
	bool pandnregmem ( long sseDestReg, void* SrcPtr );

	// vpcmpq - signed compare packed quad-word (64-bit)

	// avx512

	bool vpcmpqeq_rrm128(long kdst, long src0, void* pSrcPtr, long mask = 0);

	bool vpcmpqeq_rrb128(long kdst, long src0, long long* pSrcPtr, long mask = 0);
	bool vpcmpqeq_rrb256(long kdst, long src0, long long* pSrcPtr, long mask = 0);
	bool vpcmpqeq_rrb512(long kdst, long src0, long long* pSrcPtr, long mask = 0);

	bool vpcmpqeq_rrr128(long kdst, long src0, long src1, long mask = 0);
	bool vpcmpqne_rrr128(long kdst, long src0, long src1, long mask = 0);
	bool vpcmpqlt_rrr128(long kdst, long src0, long src1, long mask = 0);
	bool vpcmpqle_rrr128(long kdst, long src0, long src1, long mask = 0);
	bool vpcmpqgt_rrr128(long kdst, long src0, long src1, long mask = 0);
	bool vpcmpqge_rrr128(long kdst, long src0, long src1, long mask = 0);

	bool vpcmpqeq_rrr256(long kdst, long src0, long src1, long mask = 0);
	bool vpcmpqne_rrr256(long kdst, long src0, long src1, long mask = 0);
	bool vpcmpqlt_rrr256(long kdst, long src0, long src1, long mask = 0);
	bool vpcmpqle_rrr256(long kdst, long src0, long src1, long mask = 0);
	bool vpcmpqgt_rrr256(long kdst, long src0, long src1, long mask = 0);
	bool vpcmpqge_rrr256(long kdst, long src0, long src1, long mask = 0);

	bool vpcmpqeq_rrr512(long kdst, long src0, long src1, long mask = 0);
	bool vpcmpqne_rrr512(long kdst, long src0, long src1, long mask = 0);
	bool vpcmpqlt_rrr512(long kdst, long src0, long src1, long mask = 0);
	bool vpcmpqle_rrr512(long kdst, long src0, long src1, long mask = 0);
	bool vpcmpqgt_rrr512(long kdst, long src0, long src1, long mask = 0);
	bool vpcmpqge_rrr512(long kdst, long src0, long src1, long mask = 0);


	// vpcmpd - signed compare packed double-word

	// avx512

	// rrr - register-register-register
	bool vpcmpdeq_rrr128(long kdst, long src0, long src1, long mask = 0);
	bool vpcmpdne_rrr128(long kdst, long src0, long src1, long mask = 0);
	bool vpcmpdlt_rrr128(long kdst, long src0, long src1, long mask = 0);
	bool vpcmpdle_rrr128(long kdst, long src0, long src1, long mask = 0);
	bool vpcmpdgt_rrr128(long kdst, long src0, long src1, long mask = 0);
	bool vpcmpdge_rrr128(long kdst, long src0, long src1, long mask = 0);

	bool vpcmpdeq_rrr256(long kdst, long src0, long src1, long mask = 0);
	bool vpcmpdne_rrr256(long kdst, long src0, long src1, long mask = 0);
	bool vpcmpdlt_rrr256(long kdst, long src0, long src1, long mask = 0);
	bool vpcmpdle_rrr256(long kdst, long src0, long src1, long mask = 0);
	bool vpcmpdgt_rrr256(long kdst, long src0, long src1, long mask = 0);
	bool vpcmpdge_rrr256(long kdst, long src0, long src1, long mask = 0);

	bool vpcmpdeq_rrr512(long kdst, long src0, long src1, long mask = 0);
	bool vpcmpdne_rrr512(long kdst, long src0, long src1, long mask = 0);
	bool vpcmpdlt_rrr512(long kdst, long src0, long src1, long mask = 0);
	bool vpcmpdle_rrr512(long kdst, long src0, long src1, long mask = 0);
	bool vpcmpdgt_rrr512(long kdst, long src0, long src1, long mask = 0);
	bool vpcmpdge_rrr512(long kdst, long src0, long src1, long mask = 0);

	// vpcmpud - unsigned compare packed double-word

	// avx512

	// rrr - register-register-register
	bool vpcmpudeq_rrr128(long kdst, long src0, long src1, long mask = 0);
	bool vpcmpudne_rrr128(long kdst, long src0, long src1, long mask = 0);
	bool vpcmpudlt_rrr128(long kdst, long src0, long src1, long mask = 0);
	bool vpcmpudle_rrr128(long kdst, long src0, long src1, long mask = 0);
	bool vpcmpudgt_rrr128(long kdst, long src0, long src1, long mask = 0);
	bool vpcmpudge_rrr128(long kdst, long src0, long src1, long mask = 0);

	// vpcmpub - unsigned compare packed double-word

	// avx512

	// rrr - register-register-register
	bool vpcmpubeq_rrr128(long kdst, long src0, long src1, long mask = 0);
	bool vpcmpubne_rrr128(long kdst, long src0, long src1, long mask = 0);
	bool vpcmpublt_rrr128(long kdst, long src0, long src1, long mask = 0);
	bool vpcmpuble_rrr128(long kdst, long src0, long src1, long mask = 0);
	bool vpcmpubgt_rrr128(long kdst, long src0, long src1, long mask = 0);
	bool vpcmpubge_rrr128(long kdst, long src0, long src1, long mask = 0);

	// pcmpeq - compare packed integers - sets destination to all 1s if equal, sets to all 0s otherwise

	// avx
	bool pcmpeqb1regreg ( long sseDestReg, long sseSrc1Reg, long sseSrc2Reg );
	bool pcmpeqw1regreg ( long sseDestReg, long sseSrc1Reg, long sseSrc2Reg );
	bool pcmpeqd1regreg ( long sseDestReg, long sseSrc1Reg, long sseSrc2Reg );
	bool pcmpeqq1regreg(long sseDestReg, long sseSrc1Reg, long sseSrc2Reg);

	bool pcmpeqb1regmem(long sseDestReg, long sseSrc1Reg, void* pSrcPtr);
	bool pcmpeqw1regmem(long sseDestReg, long sseSrc1Reg, void* pSrcPtr);
	bool pcmpeqd1regmem(long sseDestReg, long sseSrc1Reg, void* pSrcPtr);

	bool pcmpeqb2regreg(long sseDestReg, long sseSrc1Reg, long sseSrc2Reg);
	bool pcmpeqw2regreg(long sseDestReg, long sseSrc1Reg, long sseSrc2Reg);
	bool pcmpeqd2regreg(long sseDestReg, long sseSrc1Reg, long sseSrc2Reg);
	bool pcmpeqq2regreg(long sseDestReg, long sseSrc1Reg, long sseSrc2Reg);

	// sse
	bool pcmpeqbregreg ( long sseDestReg, long sseSrcReg );
	bool pcmpeqwregreg ( long sseDestReg, long sseSrcReg );
	bool pcmpeqdregreg ( long sseDestReg, long sseSrcReg );

	bool pcmpeqbregmem ( long sseDestReg, void* SrcPtr );
	bool pcmpeqwregmem ( long sseDestReg, void* SrcPtr );
	bool pcmpeqdregmem ( long sseDestReg, void* SrcPtr );

	// pcmpgt - compare packed signed integers for greater than


	// avx
	bool pcmpgtb1regreg ( long sseDestReg, long sseSrc1Reg, long sseSrc2Reg );
	bool pcmpgtw1regreg ( long sseDestReg, long sseSrc1Reg, long sseSrc2Reg );
	bool pcmpgtd1regreg ( long sseDestReg, long sseSrc1Reg, long sseSrc2Reg );
	bool pcmpgtq1regreg(long sseDestReg, long sseSrc1Reg, long sseSrc2Reg);

	bool pcmpgtb1regmem(long sseDestReg, long sseSrc1Reg, void* pSrcPtr);
	bool pcmpgtw1regmem(long sseDestReg, long sseSrc1Reg, void* pSrcPtr);
	bool pcmpgtd1regmem(long sseDestReg, long sseSrc1Reg, void* pSrcPtr);

	bool pcmpgtb2regreg(long sseDestReg, long sseSrc1Reg, long sseSrc2Reg);
	bool pcmpgtw2regreg(long sseDestReg, long sseSrc1Reg, long sseSrc2Reg);
	bool pcmpgtd2regreg(long sseDestReg, long sseSrc1Reg, long sseSrc2Reg);
	bool pcmpgtq2regreg(long sseDestReg, long sseSrc1Reg, long sseSrc2Reg);

	// sse
	bool pcmpgtbregreg ( long sseDestReg, long sseSrcReg );
	bool pcmpgtwregreg ( long sseDestReg, long sseSrcReg );
	bool pcmpgtdregreg ( long sseDestReg, long sseSrcReg );

	bool pcmpgtbregmem ( long sseDestReg, void* SrcPtr );
	bool pcmpgtwregmem ( long sseDestReg, void* SrcPtr );
	bool pcmpgtdregmem ( long sseDestReg, void* SrcPtr );
	
	// pmaxu - get maximum of unsigned integers

	// avx512

	// rrr - register-register-register
	bool pmaxud_rrr512(long dst, long src0, long src1, long mask = 0, long z = 0);
	bool pmaxud_rrr256(long dst, long src0, long src1, long mask = 0, long z = 0);
	bool pmaxud_rrr128(long dst, long src0, long src1, long mask = 0, long z = 0);

	// avx
	bool pmaxuw1regreg(long sseDestReg, long sseSrc1Reg, long sseSrc2Reg);
	bool pmaxud1regreg(long sseDestReg, long sseSrc1Reg, long sseSrc2Reg);

	bool pmaxuw2regreg(long sseDestReg, long sseSrc1Reg, long sseSrc2Reg);
	bool pmaxud2regreg(long sseDestReg, long sseSrc1Reg, long sseSrc2Reg);

	// sse
	bool pmaxuwregreg(long sseDestReg, long sseSrcReg);
	bool pmaxudregreg(long sseDestReg, long sseSrcReg);

	bool pmaxuwregmem(long sseDestReg, void* SrcPtr);
	bool pmaxudregmem(long sseDestReg, void* SrcPtr);

	// pmaxs - get maximum of signed integers

	// avx512

	// rrr - register-register-register
	bool pmaxsd_rrr512(long dst, long src0, long src1, long mask = 0, long z = 0);
	bool pmaxsd_rrr256(long dst, long src0, long src1, long mask = 0, long z = 0);
	bool pmaxsd_rrr128(long dst, long src0, long src1, long mask = 0, long z = 0);

	// avx
	bool pmaxsw1regreg ( long sseDestReg, long sseSrc1Reg, long sseSrc2Reg );
	bool pmaxsd1regreg ( long sseDestReg, long sseSrc1Reg, long sseSrc2Reg );

	bool pmaxsw1regmem(long sseDestReg, long sseSrc1Reg, void* pSrcPtr);
	bool pmaxsd1regmem(long sseDestReg, long sseSrc1Reg, void* pSrcPtr);

	bool pmaxsw2regreg(long sseDestReg, long sseSrc1Reg, long sseSrc2Reg);
	bool pmaxsd2regreg(long sseDestReg, long sseSrc1Reg, long sseSrc2Reg);

	// sse
	bool pmaxswregreg ( long sseDestReg, long sseSrcReg );
	bool pmaxsdregreg ( long sseDestReg, long sseSrcReg );

	bool pmaxswregmem ( long sseDestReg, void* SrcPtr );
	bool pmaxsdregmem ( long sseDestReg, void* SrcPtr );

	// pminu - get minimum of unsigned integers

	// avx512

	// rrr - register-register-register
	bool pminud_rrr512(long dst, long src0, long src1, long mask = 0, long z = 0);
	bool pminud_rrr256(long dst, long src0, long src1, long mask = 0, long z = 0);
	bool pminud_rrr128(long dst, long src0, long src1, long mask = 0, long z = 0);

	// avx
	bool pminuw1regreg(long sseDestReg, long sseSrc1Reg, long sseSrc2Reg);
	bool pminud1regreg(long sseDestReg, long sseSrc1Reg, long sseSrc2Reg);

	bool pminuw1regmem(long sseDestReg, long sseSrc1Reg, void* pSrcPtr);
	bool pminud1regmem(long sseDestReg, long sseSrc1Reg, void* pSrcPtr);

	bool pminuw2regreg(long sseDestReg, long sseSrc1Reg, long sseSrc2Reg);
	bool pminud2regreg(long sseDestReg, long sseSrc1Reg, long sseSrc2Reg);

	// sse
	bool pminuwregreg(long sseDestReg, long sseSrcReg);
	bool pminudregreg(long sseDestReg, long sseSrcReg);

	bool pminuwregmem(long sseDestReg, void* SrcPtr);
	bool pminudregmem(long sseDestReg, void* SrcPtr);

	// pmins - get minimum of signed integers

	// avx512

	// rrr - register-register-register
	bool pminsd_rrr512(long dst, long src0, long src1, long mask = 0, long z = 0);
	bool pminsd_rrr256(long dst, long src0, long src1, long mask = 0, long z = 0);
	bool pminsd_rrr128(long dst, long src0, long src1, long mask = 0, long z = 0);

	// avx
	bool pminsw1regreg ( long sseDestReg, long sseSrc1Reg, long sseSrc2Reg );
	bool pminsd1regreg ( long sseDestReg, long sseSrc1Reg, long sseSrc2Reg );

	bool pminsw1regmem(long sseDestReg, long sseSrc1Reg, void* pSrcPtr);
	bool pminsd1regmem(long sseDestReg, long sseSrc1Reg, void* pSrcPtr);

	bool pminsw2regreg(long sseDestReg, long sseSrc1Reg, long sseSrc2Reg);
	bool pminsd2regreg(long sseDestReg, long sseSrc1Reg, long sseSrc2Reg);

	// sse
	bool pminswregreg ( long sseDestReg, long sseSrcReg );
	bool pminsdregreg ( long sseDestReg, long sseSrcReg );

	bool pminswregmem ( long sseDestReg, void* SrcPtr );
	bool pminsdregmem ( long sseDestReg, void* SrcPtr );
	
	// pmovsxdq - move and sign extend lower 2 32-bit packed integers into 2 64-bit packed integers

	// avx
	bool pmovsxdq1regreg ( long sseDestReg, long sseSrcReg );

	// this extends lower 4 32-bit packed integers in 128-bit register into 4 64-bit packed integers in 256-bit register
	bool pmovsxdq2regreg(long sseDestReg, long sseSrcReg);

	// sse
	bool pmovsxdqregreg ( long sseDestReg, long sseSrcReg );
	bool pmovsxdqregmem ( long sseDestReg, void* SrcPtr );
	

	// por - packed logical or of integers

	// avx512

	// rrr - register-register-register
	bool por_rrr512(long dst, long src0, long src1, long mask = 0, long z = 0);
	bool por_rrr256(long dst, long src0, long src1, long mask = 0, long z = 0);
	bool por_rrr128(long dst, long src0, long src1, long mask = 0, long z = 0);

	bool pord_rrm512(long dst, long src0, void* pSrcPtr, long mask = 0, long z = 0);
	bool pord_rrm256(long dst, long src0, void* pSrcPtr, long mask = 0, long z = 0);
	bool pord_rrm128(long dst, long src0, void* pSrcPtr, long mask = 0, long z = 0);

	bool pord_rrb512(long dst, long src0, long* pSrcPtr, long mask = 0, long z = 0);
	bool pord_rrb256(long dst, long src0, long* pSrcPtr, long mask = 0, long z = 0);
	bool pord_rrb128(long dst, long src0, long* pSrcPtr, long mask = 0, long z = 0);

	bool porq_rrr512(long dst, long src0, long src1, long mask = 0, long z = 0);
	bool porq_rrr256(long dst, long src0, long src1, long mask = 0, long z = 0);
	bool porq_rrr128(long dst, long src0, long src1, long mask = 0, long z = 0);

	bool porq_rrm512(long dst, long src0, void* pSrcPtr, long mask = 0, long z = 0);
	bool porq_rrm256(long dst, long src0, void* pSrcPtr, long mask = 0, long z = 0);
	bool porq_rrm128(long dst, long src0, void* pSrcPtr, long mask = 0, long z = 0);

	bool porq_rrb512(long dst, long src0, long long* pSrcPtr, long mask = 0, long z = 0);
	bool porq_rrb256(long dst, long src0, long long* pSrcPtr, long mask = 0, long z = 0);
	bool porq_rrb128(long dst, long src0, long long* pSrcPtr, long mask = 0, long z = 0);


	// avx
	bool por1regreg ( long sseDestReg, long sseSrc1Reg, long sseSrc2Reg );
	bool por1regmem(long sseDestReg, long sseSrc1Reg, void* pSrcPtr);

	bool por2regreg(long sseDestReg, long sseSrc1Reg, long sseSrc2Reg);

	// sse
	bool porregreg ( long sseDestReg, long sseSrcReg );
	bool porregmem ( long sseDestReg, void* SrcPtr );
	
	// pshuf - shuffle packed values

	// avx512

	// rr - register-register-immediate
	bool pshufd_rri512(long dst, long src0, char Imm8, long mask = 0, long z = 0);
	bool pshufd_rri256(long dst, long src0, char Imm8, long mask = 0, long z = 0);
	bool pshufd_rri128(long dst, long src0, char Imm8, long mask = 0, long z = 0);

	bool pshufb_rrm512(long dst, long src0, void* SrcPtr, long mask = 0, long z = 0);
	bool pshufb_rrm256(long dst, long src0, void* SrcPtr, long mask = 0, long z = 0);
	bool pshufb_rrm128(long dst, long src0, void* SrcPtr, long mask = 0, long z = 0);

	// rm - register-memory-immediate
	bool pshufd_rmi512(long dst, void* pSrcPtr, char Imm8, long mask = 0, long z = 0);
	bool pshufd_rmi256(long dst, void* pSrcPtr, char Imm8, long mask = 0, long z = 0);
	bool pshufd_rmi128(long dst, void* pSrcPtr, char Imm8, long mask = 0, long z = 0);

	// avx
	bool pshufb1regreg(long sseDestReg, long sseSrc1Reg, long sseSrc2Reg);
	bool pshufb1regmem(long sseDestReg, long sseSrc1Reg, void* SrcPtr);

	bool pshufb2regmem(long sseDestReg, long sseSrc1Reg, long x64BaseReg, long x64IndexReg, long Scale, long Offset);

	bool pshufd1regregimm ( long sseDestReg, long sseSrcReg, char Imm8 );
	bool pshufd2regregimm(long sseDestReg, long sseSrcReg, char Imm8);

	bool pshufd1regmemimm(long sseDestReg, void* SrcPtr, char Imm8);
	bool pshufd2regmemimm(long sseDestReg, void* SrcPtr, char Imm8);

	bool pshufps1regregimm(long sseDestReg, long sseSrcReg1, long sseSrcReg2, char Imm8);
	bool pshufps2regregimm(long sseDestReg, long sseSrcReg1, long sseSrcReg2, char Imm8);

	// sse
	bool pshufbregreg ( long sseDestReg, long sseSrcReg );
	bool pshufbregmem ( long sseDestReg, void* SrcPtr );
	
	bool pshufdregregimm ( long sseDestReg, long sseSrcReg, char Imm8 );
	
	// ***todo*** this instruction crashes for some reason at the moment ***needs fixing***
	bool pshufdregmemimm ( long sseDestReg, void* SrcPtr, char Imm8 );
	
	// pshufhw - packed shuffle high words

	// avx
	bool pshufhw1regregimm ( long sseDestReg, long sseSrcReg, char Imm8 );
	bool pshufhw1regmemimm(long sseDestReg, void* SrcPtr, char Imm8);

	// sse
	bool pshufhwregregimm ( long sseDestReg, long sseSrcReg, char Imm8 );
	
	// ***todo*** this instruction crashes for some reason at the moment ***needs fixing***
	bool pshufhwregmemimm ( long sseDestReg, void* SrcPtr, char Imm8 );
	
	// pshuflw - packed shuffle low words

	// avx
	bool pshuflw1regregimm ( long sseDestReg, long sseSrcReg, char Imm8 );
	bool pshuflw1regmemimm(long sseDestReg, void* SrcPtr, char Imm8);

	// sse
	bool pshuflwregregimm ( long sseDestReg, long sseSrcReg, char Imm8 );
	bool pshuflwregmemimm ( long sseDestReg, void* SrcPtr, char Imm8 );
	
	// psll - packed shift logical left integers

	// avx512

	// rri - register-register-immediate
	bool pslld_rri512(long dst, long src0, char Imm8, long mask = 0, long z = 0);
	bool pslld_rri256(long dst, long src0, char Imm8, long mask = 0, long z = 0);
	bool pslld_rri128(long dst, long src0, char Imm8, long mask = 0, long z = 0);

	bool psllq_rri512(long dst, long src0, char Imm8, long mask = 0, long z = 0);
	bool psllq_rri256(long dst, long src0, char Imm8, long mask = 0, long z = 0);
	bool psllq_rri128(long dst, long src0, char Imm8, long mask = 0, long z = 0);


	// avx
	bool pslldq1regimm ( long sseDestReg, long sseSrcReg, char Imm8 );
	bool psllw1regreg ( long sseDestReg, long sseSrc1Reg, long sseSrc2Reg );
	bool psllw1regimm ( long sseDestReg, long sseSrcReg, char Imm8 );
	bool pslld1regreg ( long sseDestReg, long sseSrc1Reg, long sseSrc2Reg );
	bool pslld1regimm ( long sseDestReg, long sseSrcReg, char Imm8 );
	bool psllq1regimm(long sseDestReg, long sseSrcReg, char Imm8);

	bool pslldq2regimm(long sseDestReg, long sseSrcReg, char Imm8);
	bool psllw2regreg(long sseDestReg, long sseSrc1Reg, long sseSrc2Reg);
	bool psllw2regimm(long sseDestReg, long sseSrcReg, char Imm8);
	bool pslld2regreg(long sseDestReg, long sseSrc1Reg, long sseSrc2Reg);
	bool pslld2regimm(long sseDestReg, long sseSrcReg, char Imm8);

	bool psllq2regimm(long sseDestReg, long sseSrcReg, char Imm8);

	// sse
	bool psllwregreg ( long sseDestReg, long sseSrcReg );
	bool psllwregmem ( long sseDestReg, void* SrcPtr );
	bool psllwregimm ( long sseDestReg, char Imm8 );

	bool pslldregreg ( long sseDestReg, long sseSrcReg );
	bool pslldregmem ( long sseDestReg, void* SrcPtr );
	bool pslldregimm ( long sseDestReg, char Imm8 );
	
	bool psllqregreg ( long sseDestReg, long sseSrcReg );
	bool psllqregmem ( long sseDestReg, void* SrcPtr );
	bool psllqregimm ( long sseDestReg, char Imm8 );

	// psra - packed shift arithemetic right integers

	// avx512

	// rri - register-register-immediate
	bool psrad_rri512(long dst, long src0, char Imm8, long mask = 0, long z = 0);
	bool psrad_rri256(long dst, long src0, char Imm8, long mask = 0, long z = 0);
	bool psrad_rri128(long dst, long src0, char Imm8, long mask = 0, long z = 0);

	bool psraq_rri512(long dst, long src0, char Imm8, long mask = 0, long z = 0);
	bool psraq_rri256(long dst, long src0, char Imm8, long mask = 0, long z = 0);
	bool psraq_rri128(long dst, long src0, char Imm8, long mask = 0, long z = 0);

	// avx2
	bool psraw1regreg ( long sseDestReg, long sseSrc1Reg, long sseSrc2Reg );
	bool psraw1regimm ( long sseDestReg, long sseSrcReg, char Imm8 );
	bool psrad1regreg ( long sseDestReg, long sseSrc1Reg, long sseSrc2Reg );
	bool psrad1regimm ( long sseDestReg, long sseSrcReg, char Imm8 );

	bool psraw2regreg(long sseDestReg, long sseSrc1Reg, long sseSrc2Reg);
	bool psraw2regimm(long sseDestReg, long sseSrcReg, char Imm8);
	bool psrad2regreg(long sseDestReg, long sseSrc1Reg, long sseSrc2Reg);
	bool psrad2regimm(long sseDestReg, long sseSrcReg, char Imm8);

	// sse
	bool psrawregreg ( long sseDestReg, long sseSrcReg );
	bool psrawregmem ( long sseDestReg, void* SrcPtr );
	bool psrawregimm ( long sseDestReg, char Imm8 );

	bool psradregreg ( long sseDestReg, long sseSrcReg );
	bool psradregmem ( long sseDestReg, void* SrcPtr );
	bool psradregimm ( long sseDestReg, char Imm8 );
	
	// psrl - packed shift logical right integers

	// avx512

	// rri - register-register-immediate
	bool psrld_rri512(long dst, long src0, char Imm8, long mask = 0, long z = 0);
	bool psrld_rri256(long dst, long src0, char Imm8, long mask = 0, long z = 0);
	bool psrld_rri128(long dst, long src0, char Imm8, long mask = 0, long z = 0);

	bool psrlq_rri512(long dst, long src0, char Imm8, long mask = 0, long z = 0);
	bool psrlq_rri256(long dst, long src0, char Imm8, long mask = 0, long z = 0);
	bool psrlq_rri128(long dst, long src0, char Imm8, long mask = 0, long z = 0);

	// avx
	bool psrldq1regimm ( long sseDestReg, long sseSrcReg, char Imm8 );
	bool psrlq1regimm(long sseDestReg, long sseSrcReg, char Imm8);
	bool psrlw1regreg ( long sseDestReg, long sseSrc1Reg, long sseSrc2Reg );
	bool psrlw1regimm ( long sseDestReg, long sseSrcReg, char Imm8 );
	bool psrld1regreg ( long sseDestReg, long sseSrc1Reg, long sseSrc2Reg );
	bool psrld1regimm ( long sseDestReg, long sseSrcReg, char Imm8 );

	bool psrldq2regimm(long sseDestReg, long sseSrcReg, char Imm8);
	bool psrlq2regimm(long sseDestReg, long sseSrcReg, char Imm8);
	bool psrlw2regreg(long sseDestReg, long sseSrc1Reg, long sseSrc2Reg);
	bool psrlw2regimm(long sseDestReg, long sseSrcReg, char Imm8);
	bool psrld2regreg(long sseDestReg, long sseSrc1Reg, long sseSrc2Reg);
	bool psrld2regimm(long sseDestReg, long sseSrcReg, char Imm8);

	// sse
	bool psrlwregreg ( long sseDestReg, long sseSrcReg );
	bool psrlwregmem ( long sseDestReg, void* SrcPtr );
	bool psrlwregimm ( long sseDestReg, char Imm8 );

	bool psrldregreg ( long sseDestReg, long sseSrcReg );
	bool psrldregmem ( long sseDestReg, void* SrcPtr );
	bool psrldregimm ( long sseDestReg, char Imm8 );
	
	bool psrlqregreg ( long sseDestReg, long sseSrcReg );
	bool psrlqregmem ( long sseDestReg, void* SrcPtr );
	bool psrlqregimm ( long sseDestReg, char Imm8 );
	
	// psub - subtraction of packed integers

	// avx512

	// rrm - register-register-memory
	bool psubd_rrm512(long dst, long src0, void* pSrcPtr, long mask = 0, long z = 0);
	bool psubd_rrm256(long dst, long src0, void* pSrcPtr, long mask = 0, long z = 0);
	bool psubd_rrm128(long dst, long src0, void* pSrcPtr, long mask = 0, long z = 0);

	bool psubq_rrm512(long dst, long src0, void* pSrcPtr, long mask = 0, long z = 0);
	bool psubq_rrm256(long dst, long src0, void* pSrcPtr, long mask = 0, long z = 0);
	bool psubq_rrm128(long dst, long src0, void* pSrcPtr, long mask = 0, long z = 0);


	// rrb - register-register-broadcast memory
	bool psubd_rrb512(long dst, long src0, long* pSrcPtr, long mask = 0, long z = 0);
	bool psubd_rrb256(long dst, long src0, long* pSrcPtr, long mask = 0, long z = 0);
	bool psubd_rrb128(long dst, long src0, long* pSrcPtr, long mask = 0, long z = 0);

	bool psubq_rrb512(long dst, long src0, long long* pSrcPtr, long mask = 0, long z = 0);
	bool psubq_rrb256(long dst, long src0, long long* pSrcPtr, long mask = 0, long z = 0);
	bool psubq_rrb128(long dst, long src0, long long* pSrcPtr, long mask = 0, long z = 0);

	// rrr - register-register-register
	bool psubd_rrr512(long dst, long src0, long src1, long mask = 0, long z = 0);
	bool psubd_rrr256(long dst, long src0, long src1, long mask = 0, long z = 0);
	bool psubd_rrr128(long dst, long src0, long src1, long mask = 0, long z = 0);

	bool psubq_rrr512(long dst, long src0, long src1, long mask = 0, long z = 0);
	bool psubq_rrr256(long dst, long src0, long src1, long mask = 0, long z = 0);
	bool psubq_rrr128(long dst, long src0, long src1, long mask = 0, long z = 0);

	// avx
	bool psubb1regreg ( long sseDestReg, long sseSrc1Reg, long sseSrc2Reg );
	bool psubw1regreg ( long sseDestReg, long sseSrc1Reg, long sseSrc2Reg );
	bool psubd1regreg ( long sseDestReg, long sseSrc1Reg, long sseSrc2Reg );
	bool psubq1regreg(long sseDestReg, long sseSrc1Reg, long sseSrc2Reg);

	bool psubb1regmem(long sseDestReg, long sseSrc1Reg, void* pSrcPtr);
	bool psubw1regmem(long sseDestReg, long sseSrc1Reg, void* pSrcPtr);
	bool psubd1regmem(long sseDestReg, long sseSrc1Reg, void* pSrcPtr);

	bool psubb2regreg(long sseDestReg, long sseSrc1Reg, long sseSrc2Reg);
	bool psubw2regreg(long sseDestReg, long sseSrc1Reg, long sseSrc2Reg);
	bool psubd2regreg(long sseDestReg, long sseSrc1Reg, long sseSrc2Reg);
	bool psubq2regreg(long sseDestReg, long sseSrc1Reg, long sseSrc2Reg);

	// sse
	bool psubbregreg ( long sseDestReg, long sseSrcReg );
	bool psubwregreg ( long sseDestReg, long sseSrcReg );
	bool psubdregreg ( long sseDestReg, long sseSrcReg );
	bool psubqregreg ( long sseDestReg, long sseSrcReg );

	bool psubbregmem ( long sseDestReg, void* SrcPtr );
	bool psubwregmem ( long sseDestReg, void* SrcPtr );
	bool psubdregmem ( long sseDestReg, void* SrcPtr );
	bool psubqregmem ( long sseDestReg, void* SrcPtr );
	
	// psubs - subtract packed integers with signed saturation

	// avx
	bool psubsb1regreg ( long sseDestReg, long sseSrc1Reg, long sseSrc2Reg );
	bool psubsw1regreg ( long sseDestReg, long sseSrc1Reg, long sseSrc2Reg );
	
	bool psubsb1regmem(long sseDestReg, long sseSrc1Reg, void* pSrcPtr);
	bool psubsw1regmem(long sseDestReg, long sseSrc1Reg, void* pSrcPtr);


	// psubus - subtract packed integers with unsigned saturation

	// avx
	bool psubusb1regreg ( long sseDestReg, long sseSrc1Reg, long sseSrc2Reg );
	bool psubusw1regreg ( long sseDestReg, long sseSrc1Reg, long sseSrc2Reg );

	bool psubusb1regmem(long sseDestReg, long sseSrc1Reg, void* pSrcPtr);
	bool psubusw1regmem(long sseDestReg, long sseSrc1Reg, void* pSrcPtr);


	// sse
	bool psubsbregreg ( long sseDestReg, long sseSrcReg );
	bool psubsbregmem ( long sseDestReg, void* SrcPtr );
	bool psubswregreg ( long sseDestReg, long sseSrcReg );
	bool psubswregmem ( long sseDestReg, void* SrcPtr );
	
	bool psubusbregreg ( long sseDestReg, long sseSrcReg );
	bool psubusbregmem ( long sseDestReg, void* SrcPtr );
	bool psubuswregreg ( long sseDestReg, long sseSrcReg );
	bool psubuswregmem ( long sseDestReg, void* SrcPtr );
	
	// punpckh - unpack from high - first source on bottom

	// avx
	bool punpckhbw1regreg ( long sseDestReg, long sseSrc1Reg, long sseSrc2Reg );
	bool punpckhwd1regreg ( long sseDestReg, long sseSrc1Reg, long sseSrc2Reg );
	bool punpckhdq1regreg ( long sseDestReg, long sseSrc1Reg, long sseSrc2Reg );
	bool punpckhqdq1regreg ( long sseDestReg, long sseSrc1Reg, long sseSrc2Reg );

	bool punpckhbw1regmem(long sseDestReg, long sseSrc1Reg, void* pSrcPtr);
	bool punpckhwd1regmem(long sseDestReg, long sseSrc1Reg, void* pSrcPtr);
	bool punpckhdq1regmem(long sseDestReg, long sseSrc1Reg, void* pSrcPtr);
	bool punpckhqdq1regmem(long sseDestReg, long sseSrc1Reg, void* pSrcPtr);

	// sse
	bool punpckhbwregreg ( long sseDestReg, long sseSrcReg );
	bool punpckhbwregmem ( long sseDestReg, void* SrcPtr );
	
	bool punpckhwdregreg ( long sseDestReg, long sseSrcReg );
	bool punpckhwdregmem ( long sseDestReg, void* SrcPtr );

	bool punpckhdqregreg ( long sseDestReg, long sseSrcReg );
	bool punpckhdqregmem ( long sseDestReg, void* SrcPtr );
	
	bool punpckhqdqregreg ( long sseDestReg, long sseSrcReg );
	bool punpckhqdqregmem ( long sseDestReg, void* SrcPtr );
	
	// punpckl - unpack from low - first source on bottom


	// avx
	bool punpcklbw1regreg ( long sseDestReg, long sseSrc1Reg, long sseSrc2Reg );
	bool punpcklwd1regreg ( long sseDestReg, long sseSrc1Reg, long sseSrc2Reg );
	bool punpckldq1regreg ( long sseDestReg, long sseSrc1Reg, long sseSrc2Reg );
	bool punpcklqdq1regreg ( long sseDestReg, long sseSrc1Reg, long sseSrc2Reg );

	bool punpcklbw1regmem(long sseDestReg, long sseSrc1Reg, void* pSrcPtr);
	bool punpcklwd1regmem(long sseDestReg, long sseSrc1Reg, void* pSrcPtr);
	bool punpckldq1regmem(long sseDestReg, long sseSrc1Reg, void* pSrcPtr);
	bool punpcklqdq1regmem(long sseDestReg, long sseSrc1Reg, void* pSrcPtr);

	// sse
	bool punpcklbwregreg ( long sseDestReg, long sseSrcReg );
	bool punpcklbwregmem ( long sseDestReg, void* SrcPtr );
	
	bool punpcklwdregreg ( long sseDestReg, long sseSrcReg );
	bool punpcklwdregmem ( long sseDestReg, void* SrcPtr );

	bool punpckldqregreg ( long sseDestReg, long sseSrcReg );
	bool punpckldqregmem ( long sseDestReg, void* SrcPtr );
	
	bool punpcklqdqregreg ( long sseDestReg, long sseSrcReg );
	bool punpcklqdqregmem ( long sseDestReg, void* SrcPtr );


	// pxor

	// avx512

	// rrr - register-register-register
	bool pxor_rrr512(long dst, long src0, long src1, long mask = 0, long z = 0);
	bool pxor_rrr256(long dst, long src0, long src1, long mask = 0, long z = 0);
	bool pxor_rrr128(long dst, long src0, long src1, long mask = 0, long z = 0);

	// avx
	bool pxor1regreg ( long sseDestReg, long sseSrc1Reg, long sseSrc2Reg );
	bool pxor1regmem(long sseDestReg, long sseSrc1Reg, void* pSrcPtr);

	bool pxor2regreg(long sseDestReg, long sseSrc1Reg, long sseSrc2Reg);


	// sse
	bool pxorregreg ( long sseDestReg, long sseSrcReg );
	bool pxorregmem ( long sseDestReg, void* SrcPtr );

	// vpmov b/w/d/q 2m - convert vector register to mask register

	bool vpmovb2m_rr512(long dst, long src0);
	bool vpmovb2m_rr256(long dst, long src0);
	bool vpmovb2m_rr128(long dst, long src0);

	bool vpmovd2m_rr512(long dst, long src0);
	bool vpmovd2m_rr256(long dst, long src0);
	bool vpmovd2m_rr128(long dst, long src0);


	// vpmovm2b/w/d/q - convert mask register to vector register

	bool vpmovm2b_rr512(long dst, long src0);
	bool vpmovm2b_rr256(long dst, long src0);
	bool vpmovm2b_rr128(long dst, long src0);

	bool vpmovm2w_rr512(long dst, long src0);
	bool vpmovm2w_rr256(long dst, long src0);
	bool vpmovm2w_rr128(long dst, long src0);

	bool vpmovm2d_rr512(long dst, long src0);
	bool vpmovm2d_rr256(long dst, long src0);
	bool vpmovm2d_rr128(long dst, long src0);

	bool vpmovm2q_rr512(long dst, long src0);
	bool vpmovm2q_rr256(long dst, long src0);
	bool vpmovm2q_rr128(long dst, long src0);


	// vptestmb/w/d/q - do logical AND and set bit in mask if result is non-zero

	// avx512

	// rr - register-register
	bool vptestmb_rrr512(long kdst, long src0, long src1, long mask = 0);
	bool vptestmb_rrr256(long kdst, long src0, long src1, long mask = 0);
	bool vptestmb_rrr128(long kdst, long src0, long src1, long mask = 0);

	bool vptestmw_rrr512(long kdst, long src0, long src1, long mask = 0);
	bool vptestmw_rrr256(long kdst, long src0, long src1, long mask = 0);
	bool vptestmw_rrr128(long kdst, long src0, long src1, long mask = 0);

	bool vptestmd_rrr512(long kdst, long src0, long src1, long mask = 0);
	bool vptestmd_rrr256(long kdst, long src0, long src1, long mask = 0);
	bool vptestmd_rrr128(long kdst, long src0, long src1, long mask = 0);

	bool vptestmd_rrm512(long kdst, long src0, void* pSrcPtr, long mask = 0);
	bool vptestmd_rrm256(long kdst, long src0, void* pSrcPtr, long mask = 0);
	bool vptestmd_rrm128(long kdst, long src0, void* pSrcPtr, long mask = 0);

	bool vptestmd_rrb512(long kdst, long src0, long* pSrcPtr32, long mask = 0);
	bool vptestmd_rrb256(long kdst, long src0, long* pSrcPtr32, long mask = 0);
	bool vptestmd_rrb128(long kdst, long src0, long* pSrcPtr32, long mask = 0);

	bool vptestmq_rrr512(long kdst, long src0, long src1, long mask = 0);
	bool vptestmq_rrr256(long kdst, long src0, long src1, long mask = 0);
	bool vptestmq_rrr128(long kdst, long src0, long src1, long mask = 0);

	bool vptestmq_rrm512(long dst, long src0, void* pSrcPtr, long mask = 0);
	bool vptestmq_rrm256(long dst, long src0, void* pSrcPtr, long mask = 0);
	bool vptestmq_rrm128(long dst, long src0, void* pSrcPtr, long mask = 0);

	bool vptestmq_rrb512(long dst, long src0, long long* pSrcPtr64, long mask = 0);
	bool vptestmq_rrb256(long dst, long src0, long long* pSrcPtr64, long mask = 0);
	bool vptestmq_rrb128(long dst, long src0, long long* pSrcPtr64, long mask = 0);

	// vptestnmb/w/d/q - do logical AND and set bit in mask if result is non-zero

	bool vptestnmb_rrr512(long kdst, long src0, long src1, long mask = 0);
	bool vptestnmb_rrr256(long kdst, long src0, long src1, long mask = 0);
	bool vptestnmb_rrr128(long kdst, long src0, long src1, long mask = 0);

	bool vptestnmw_rrr512(long kdst, long src0, long src1, long mask = 0);
	bool vptestnmw_rrr256(long kdst, long src0, long src1, long mask = 0);
	bool vptestnmw_rrr128(long kdst, long src0, long src1, long mask = 0);

	bool vptestnmd_rrr512(long kdst, long src0, long src1, long mask = 0);
	bool vptestnmd_rrr256(long kdst, long src0, long src1, long mask = 0);
	bool vptestnmd_rrr128(long kdst, long src0, long src1, long mask = 0);

	bool vptestnmd_rrm512(long dst, long src0, void* pSrcPtr, long mask = 0);
	bool vptestnmd_rrm256(long dst, long src0, void* pSrcPtr, long mask = 0);
	bool vptestnmd_rrm128(long dst, long src0, void* pSrcPtr, long mask = 0);

	bool vptestnmq_rrr512(long kdst, long src0, long src1, long mask = 0);
	bool vptestnmq_rrr256(long kdst, long src0, long src1, long mask = 0);
	bool vptestnmq_rrr128(long kdst, long src0, long src1, long mask = 0);

	bool vptestnmq_rrm512(long dst, long src0, void* pSrcPtr, long mask = 0);
	bool vptestnmq_rrm256(long dst, long src0, void* pSrcPtr, long mask = 0);
	bool vptestnmq_rrm128(long dst, long src0, void* pSrcPtr, long mask = 0);


	// ptest

	// avx
	bool ptest1regreg(long sseDestReg, long sseSrcReg);
	bool ptest2regreg(long sseDestReg, long sseSrcReg);

	// sse
	bool ptestregreg ( long sseDestReg, long sseSrcReg );
	bool ptestregmem ( long sseDestReg, void* SrcPtr );

	
	// pmuldq - multiplies 2 signed 32-bit values (index 0 and 2) and stores the results as two 64-bit values

	// sse
	bool pmuldqregreg ( long sseDestReg, long sseSrcReg );
	bool pmuldqregmem ( long sseDestReg, void* SrcPtr );

	// avx
	bool pmuldq1regreg(long sseDestReg, long sseSrc1Reg, long sseSrc2Reg);
	bool pmuldq1regmem(long sseDestReg, long sseSrc1Reg, void* SrcPtr);

	// pmuludq - unsigned version of pmuldq

	// avx
	bool pmuludq1regreg(long sseDestReg, long sseSrc1Reg, long sseSrc2Reg);
	bool pmuludq1regmem(long sseDestReg, long sseSrc1Reg, void* SrcPtr);

	bool pmuludq2regreg(long sseDestReg, long sseSrc1Reg, long sseSrc2Reg);

	// sse
	bool pmuludqregreg ( long sseDestReg, long sseSrcReg );
	bool pmuludqregmem ( long sseDestReg, void* SrcPtr );
	
	// pmulld - multiplies 4 signed 32-bit values and stores low result
	bool pmulldregreg ( long sseDestReg, long sseSrcReg );
	bool pmulldregmem ( long sseDestReg, void* SrcPtr );
	
	
	// pmullw - multiply packed signed 16-bit halfwords and store low result

	// sse
	bool pmullwregreg ( long sseDestReg, long sseSrcReg );
	bool pmullwregmem ( long sseDestReg, void* SrcPtr );

	// avx
	bool pmullw1regreg(long sseDestReg, long sseSrc1Reg, long sseSrc2Reg);

	// pmulhw - multiply packed signed 16-bit halfwords and store high result

	// sse
	bool pmulhwregreg ( long sseDestReg, long sseSrcReg );
	bool pmulhwregmem ( long sseDestReg, void* SrcPtr );
	
	// avx
	bool pmulhw1regreg(long sseDestReg, long sseSrc1Reg, long sseSrc2Reg);

	// pmaddwd - mulitply halfwords and add adjacent results

	// sse
	bool pmaddwdregreg ( long sseDestReg, long sseSrcReg );
	bool pmaddwdregmem ( long sseDestReg, void* SrcPtr );

	// avx
	bool pmaddwd1regreg(long sseDestReg, long sseSrc1Reg, long sseSrc2Reg);
	bool pmaddwd1regmem(long sseDestReg, long sseSrc1Reg, void* pSrcPtr);

	// psign - make result zero if arg zero, negate result if arg negative, otherwise leave result alone

	// avx
	bool psignb1regreg(long sseDestReg, long sseSrc1Reg, long sseSrc2Reg);
	bool psignw1regreg(long sseDestReg, long sseSrc1Reg, long sseSrc2Reg);
	bool psignd1regreg(long sseDestReg, long sseSrc1Reg, long sseSrc2Reg);

	bool psignb2regreg(long sseDestReg, long sseSrc1Reg, long sseSrc2Reg);
	bool psignw2regreg(long sseDestReg, long sseSrc1Reg, long sseSrc2Reg);
	bool psignd2regreg(long sseDestReg, long sseSrc1Reg, long sseSrc2Reg);

	// sse
	bool psignbregreg (long sseDestReg, long sseSrcReg);
	bool psignwregreg(long sseDestReg, long sseSrcReg);
	bool psigndregreg(long sseDestReg, long sseSrcReg);


	// vpshiftv - variable shift

	// vpsllv - variable shift left logical

	// avx
	bool vpsllvd1regreg(long sseDestReg, long sseSrc1Reg, long sseSrc2Reg);
	bool vpsllvd2regreg(long sseDestReg, long sseSrc1Reg, long sseSrc2Reg);


	// vpsrav - variable shift right arithmetic

	// avx
	bool vpsravd1regreg(long sseDestReg, long sseSrc1Reg, long sseSrc2Reg);
	bool vpsravd2regreg(long sseDestReg, long sseSrc1Reg, long sseSrc2Reg);


	// vpsrlv - variable shift right logical

	// avx
	bool vpsrlvd1regreg(long sseDestReg, long sseSrc1Reg, long sseSrc2Reg);
	bool vpsrlvd2regreg(long sseDestReg, long sseSrc1Reg, long sseSrc2Reg);


	//** SSE Floating Point Instructions **//
	
	// extractps
	bool extractps ( long x64DestReg, long sseSrcReg, char Imm8 );

	// movmskb

	// avx
	bool movmskb1regreg(long x64DestReg, long sseSrcReg);
	bool movmskb2regreg(long x64DestReg, long sseSrcReg);

	// sse
	bool movmskbregreg(long x64DestReg, long sseSrcReg);


	// movmskps

	// avx
	bool movmskps1regreg(long x64DestReg, long sseSrcReg);
	bool movmskps2regreg ( long x64DestReg, long sseSrcReg );
	
	// sse
	bool movmskpsregreg ( long x64DestReg, long sseSrcReg );

	// movmskpd

	// avx
	bool movmskpd1regreg(long x64DestReg, long sseSrcReg);
	bool movmskpd2regreg(long x64DestReg, long sseSrcReg);


	// vcvtdq2ps

	// avx512
	bool cvtdq2ps_rr512(long dst, long src0, long mask = 0, long z = 0);
	bool cvtdq2ps_rr256(long dst, long src0, long mask = 0, long z = 0);
	bool cvtdq2ps_rr128(long dst, long src0, long mask = 0, long z = 0);

	// avx
	bool cvtdq2ps1regreg(long sseDestReg, long sseSrcReg);
	bool cvtdq2ps2regreg(long sseDestReg, long sseSrcReg);

	bool cvtsi2ss1regreg(long sseDestReg, long sseSrcReg, long x64SrcReg);
	bool cvtsi2ss1regmem(long sseDestReg, long sseSrcReg, long* pSrcPtr);

	// sse
	bool cvtdq2psregreg(long sseDestReg, long sseSrcReg);


	// cvtsd2ss

	// avx
	bool cvtsd2ss1regreg(long sseDestReg, long sseSrcReg);

	
	// movaps
	bool movaps ( long sseDestReg, long sseSrcReg );
	bool movapd ( long sseDestReg, long sseSrcReg );
	bool movapstomem ( long sseSrcReg, long x64BaseReg, long x64IndexReg, long Scale, long Offset );
	bool movapsfrommem ( long sseDestReg, long x64BaseReg, long x64IndexReg, long Scale, long Offset );
	bool movapdtomem ( long sseSrcReg, long x64BaseReg, long x64IndexReg, long Scale, long Offset );
	bool movapdfrommem ( long sseDestReg, long x64BaseReg, long x64IndexReg, long Scale, long Offset );
	

	// vbroadcastss
	bool vbroadcastss ( long sseDestReg, long x64BaseReg, long x64IndexReg, long Scale, long Offset );
	bool vbroadcastsd ( long sseDestReg, long x64BaseReg, long x64IndexReg, long Scale, long Offset );
	
	// vmaskmov
	bool vmaskmovpstomem ( long sseSrcReg, long x64BaseReg, long x64IndexReg, long Scale, long Offset, long sseMaskReg );
	bool vmaskmovpdtomem ( long sseSrcReg, long x64BaseReg, long x64IndexReg, long Scale, long Offset, long sseMaskReg );

	bool vmaskmovd1_memreg(long sseMaskReg,long sseSrcReg, long x64BaseReg, long x64IndexReg, long Scale, long Offset);
	bool vmaskmovd1_memreg(void* pDestPtr, long sseSrc1Reg, long sseSrc2Reg);
	bool vmaskmovq1_memreg(void* pDestPtr, long sseSrc1Reg, long sseSrc2Reg);
	bool vmaskmovq1_memreg(long sseMaskReg, long sseSrcReg, long x64BaseReg, long x64IndexReg, long Scale, long Offset);

	bool vmaskmovd2_memreg(long sseMaskReg, long sseSrcReg, long x64BaseReg, long x64IndexReg, long Scale, long Offset);
	bool vmaskmovd2_memreg(void* pDestPtr, long sseSrc1Reg, long sseSrc2Reg);
	bool vmaskmovq2_memreg(void* pDestPtr, long sseSrc1Reg, long sseSrc2Reg);
	bool vmaskmovq2_memreg(long sseMaskReg, long sseSrcReg, long x64BaseReg, long x64IndexReg, long Scale, long Offset);

	// blendvp

	// avx
	bool blendvps1regreg ( long sseDestReg, long sseSrc1Reg, long sseSrc2Reg, long sseSrc3Reg );
	bool blendvpd1regreg ( long sseDestReg, long sseSrc1Reg, long sseSrc2Reg, long sseSrc3Reg );

	bool blendvps2regreg(long sseDestReg, long sseSrc1Reg, long sseSrc2Reg, long sseSrc3Reg);
	bool blendvpd2regreg(long sseDestReg, long sseSrc1Reg, long sseSrc2Reg, long sseSrc3Reg);

	bool blendps ( long sseDestReg, long sseSrc1Reg, long sseSrc2Reg, char Imm8 );
	bool blendpd ( long sseDestReg, long sseSrc1Reg, long sseSrc2Reg, char Imm8 );

	bool blendvps ( long sseDestReg, long sseSrc1Reg, long x64BaseReg, long x64IndexReg, long Scale, long Offset, long sseSrc3Reg );
	bool blendvpd ( long sseDestReg, long sseSrc1Reg, long x64BaseReg, long x64IndexReg, long Scale, long Offset, long sseSrc3Reg );
	bool blendps ( long sseDestReg, long sseSrc1Reg, long x64BaseReg, long x64IndexReg, long Scale, long Offset, char Imm8 );
	bool blendpd ( long sseDestReg, long sseSrc1Reg, long x64BaseReg, long x64IndexReg, long Scale, long Offset, char Imm8 );


	// vpmovsxbd - up convert 4/8 bytes into mask for xmm/ymm

	// avx2
	bool vpmovsxbd1regreg(long sseDestReg, long sseSrcReg);
	bool vpmovsxbd2regreg(long sseDestReg, long sseSrcReg);

	bool vpmovsxbd1regmem(long sseDestReg, long* SrcPtr);
	bool vpmovsxbd2regmem(long sseDestReg, long long* SrcPtr);

	// vpmovsxdq - up convert 32-bit value to 64-bit value with sign-extension

	// avx512

	// rr - register-register
	bool vpmovsxdq_rr512(long dst, long src0, long mask = 0, long z = 0);
	bool vpmovsxdq_rr256(long dst, long src0, long mask = 0, long z = 0);
	bool vpmovsxdq_rr128(long dst, long src0, long mask = 0, long z = 0);

	// vpmovzxdq - up convert 32-bit value to 64-bit value with zero-extension

	// avx512

	// rr - register-register
	bool vpmovzxdq_rr512(long dst, long src0, long mask = 0, long z = 0);
	bool vpmovzxdq_rr256(long dst, long src0, long mask = 0, long z = 0);
	bool vpmovzxdq_rr128(long dst, long src0, long mask = 0, long z = 0);

	// vpmovqd - down convert 64-bit value to 32-bit value with truncation

	// avx512

	// rr - register-register
	bool vpmovqd_rr512(long dst, long src0, long mask = 0, long z = 0);
	bool vpmovqd_rr256(long dst, long src0, long mask = 0, long z = 0);
	bool vpmovqd_rr128(long dst, long src0, long mask = 0, long z = 0);


	// vpbroadcastd

	// avx2
	bool vpbroadcastd1regreg(long sseDstReg, long sseSrcReg);
	bool vpbroadcastd1regmem(long sseDstReg, long* SrcPtr);
	bool vpbroadcastd2regreg(long sseDstReg, long sseSrcReg);
	bool vpbroadcastd2regmem(long sseDstReg, long* SrcPtr);

	// avx512

	// rr - register-register
	bool vpbroadcastd_rx512(long dst, long x64reg, long mask = 0, long z = 0);
	bool vpbroadcastd_rx256(long dst, long x64reg, long mask = 0, long z = 0);
	bool vpbroadcastd_rx128(long dst, long x64reg, long mask = 0, long z = 0);

	bool vpbroadcastd_rm512(long dst, long* SrcPtr, long mask = 0, long z = 0);
	bool vpbroadcastd_rm256(long dst, long* SrcPtr, long mask = 0, long z = 0);
	bool vpbroadcastd_rm128(long dst, long* SrcPtr, long mask = 0, long z = 0);

	// vpbroadcastq

	// avx2
	bool vpbroadcastq1regreg(long sseDstReg, long sseSrcReg);
	bool vpbroadcastq1regmem(long sseDstReg, long long* SrcPtr);
	bool vpbroadcastq2regreg(long sseDstReg, long sseSrcReg);
	bool vpbroadcastq2regmem(long sseDstReg, long long* SrcPtr);

	// avx512

	// rr - register-register
	bool vpbroadcastq_rr512(long dst, long x64reg, long mask = 0, long z = 0);
	bool vpbroadcastq_rr256(long dst, long x64reg, long mask = 0, long z = 0);
	bool vpbroadcastq_rr128(long dst, long x64reg, long mask = 0, long z = 0);


	// vpternlogd - 32-bit ternary instructions

	// avx512

	bool vpternlogd_rrr512(long dst, long src0, long src1, char Imm8, long mask = 0, long z = 0);
	bool vpternlogd_rrr256(long dst, long src0, long src1, char Imm8, long mask = 0, long z = 0);
	bool vpternlogd_rrr128(long dst, long src0, long src1, char Imm8, long mask = 0, long z = 0);

	bool vpternlogq_rrr512(long dst, long src0, long src1, char Imm8, long mask = 0, long z = 0);
	bool vpternlogq_rrr256(long dst, long src0, long src1, char Imm8, long mask = 0, long z = 0);
	bool vpternlogq_rrr128(long dst, long src0, long src1, char Imm8, long mask = 0, long z = 0);


	// *** SSE instructions *** //

	// avx
	bool movd1_regmem(long sseDestReg, long* SrcPtr);
	bool movd1_memreg(long* DstPtr, long sseSrcReg);
	bool movq1_regmem(long sseDestReg, long long* SrcPtr);
	bool movq1_memreg(long long* DstPtr, long sseSrcReg);

	//bool movd1_memreg(long sseSrcReg, long x64BaseReg, long x64IndexReg, long Scale, long Offset);
	//bool movd1_regmem(long sseDestReg, long x64BaseReg, long x64IndexReg, long Scale, long Offset);
	//bool movq1_memreg(long sseSrcReg, long x64BaseReg, long x64IndexReg, long Scale, long Offset);
	//bool movq1_regmem(long sseDestReg, long x64BaseReg, long x64IndexReg, long Scale, long Offset);

	bool movd_to_sse128(long sseDestReg, long x64SrcReg);
	bool movd_from_sse128(long x64DestReg, long sseSrcReg);

	bool movq_to_sse128(long sseDestReg, long x64SrcReg);
	bool movq_from_sse128(long x64DestReg, long sseSrcReg);

	// avx512
	bool movd_to_sse_rr128(long dst, long x64reg);
	bool movd_to_x64_rr128(long x64reg, long src0);
	bool movd_to_sse_rm128(long dst, long* pSrcPtr);
	bool movd_to_x64_mr128(long* pDstPtr, long src0);

	bool movq_to_sse_rr128(long dst, long x64reg);
	bool movq_to_x64_rr128(long x64reg, long src0);
	bool movq_to_sse_rm128(long dst, long long* pSrcPtr);
	bool movq_to_x64_mr128(long long* pDstPtr, long src0);


	// sse
	bool movd_to_sse ( long sseDestReg, long x64SrcReg );
	bool movd_from_sse ( long x64DestReg, long sseSrcReg );
	
	bool movq_to_sse ( long sseDestReg, long x64SrcReg );
	bool movq_from_sse ( long x64DestReg, long sseSrcReg );

	bool movd_regmem ( long sseDestReg, long* SrcPtr );
	bool movd_memreg ( long* DstPtr, long sseSrcReg );
	bool movq_regmem ( long sseDestReg, long* SrcPtr );
	bool movq_memreg ( long* DstPtr, long sseSrcReg );


	
	// move and duplicate double floating point values
	bool movddup_regreg ( long sseDestReg, long sseSrcReg );
	bool movddup_regmem ( long sseDestReg, long long* SrcPtr );
	
	

	
	// *** SSE floating point instructions *** //
	

	bool addsd ( long sseDestReg, long sseSrcReg );
	bool subsd ( long sseDestReg, long sseSrcReg );
	bool mulsd ( long sseDestReg, long sseSrcReg );
	bool divsd ( long sseDestReg, long sseSrcReg );
	bool sqrtsd ( long sseDestReg, long sseSrcReg );
	
	bool mulss(long sseDestReg, long sseSrcReg);

	bool addpsregreg(long sseDestReg, long sseSrcReg);
	bool mulpsregreg(long sseDestReg, long sseSrcReg);

	bool addpdregreg ( long sseDestReg, long sseSrcReg );
	bool addpdregmem ( long sseDestReg, void* SrcPtr );

	bool mulpdregreg ( long sseDestReg, long sseSrcReg );
	bool mulpdregmem ( long sseDestReg, void* SrcPtr );
	
	// convert with truncation scalar single precision floating point value to general purpose register signed doubleword

	// avx512
	bool cvttps2dq_rr512(long dst, long src0, long mask = 0, long z = 0);
	bool cvttps2dq_rr256(long dst, long src0, long mask = 0, long z = 0);
	bool cvttps2dq_rr128(long dst, long src0, long mask = 0, long z = 0);

	bool cvttss2si_rr128(long dst, long src0);

	// avx
	bool cvttps2dq1regreg(long sseDestReg, long sseSrcReg);
	bool cvttps2dq2regreg(long sseDestReg, long sseSrcReg);

	bool cvttss2si1regreg(long sseDestReg, long sseSrcReg);

	bool cvtss2sd1regreg(long sseDestReg, long sseSrcReg);

	// sse
	bool cvttss2si ( long x64DestReg, long sseSrcReg );
	bool cvttps2dq_regreg ( long sseDestReg, long sseSrcReg );
	
	// convert doubleword integer to scalar double precision floating point value
	bool cvtsi2sd ( long sseDestReg, long x64SrcReg );
	bool cvtdq2pd_regreg ( long sseDestReg, long sseSrcReg );

	bool cvtss2sdregreg(long sseDestReg, long sseSrcReg);

	// avx 512

	bool cvtss2sd_rr32(long dst, long src0, long mask = 0, long z = 0);
	bool cvtss2sd_rm32(long avxDstReg, long* pSrcPtr, long mask = 0, long z = 0);

	bool cvtsd2ss_rr64(long dst, long src0, long mask = 0, long z = 0);

	bool cvtps2pd_rr512(long dst, long src0, long mask = 0, long z = 0);
	bool cvtps2pd_rr256(long dst, long src0, long mask = 0, long z = 0);
	bool cvtps2pd_rr128(long dst, long src0, long mask = 0, long z = 0);

	bool cvtpd2ps_rr512(long dst, long src0, long mask = 0, long z = 0);
	bool cvtpd2ps_rr256(long dst, long src0, long mask = 0, long z = 0);
	bool cvtpd2ps_rr128(long dst, long src0, long mask = 0, long z = 0);


	// *** AVX floating point instructions *** //
	
	// addp

	// avx512
	bool vaddpd_rrr512(long dst, long src0, long src1, long mask = 0, long z = 0);
	bool vaddpd_rrr256(long dst, long src0, long src1, long mask = 0, long z = 0);
	bool vaddpd_rrr128(long dst, long src0, long src1, long mask = 0, long z = 0);

	bool vaddps_rrr512(long dst, long src0, long src1, long mask = 0, long z = 0, long rnd_enable = 0, long rnd_type = 0);
	bool vaddps_rrr256(long dst, long src0, long src1, long mask = 0, long z = 0);
	bool vaddps_rrr128(long dst, long src0, long src1, long mask = 0, long z = 0);

	bool vaddss_rrr128(long dst, long src0, long src1, long mask = 0, long z = 0, long rnd_enable = 0, long rnd_type = 0);
	bool vaddsd_rrr128(long dst, long src0, long src1, long mask = 0, long z = 0);

	// avx

	bool vaddps ( long sseDestReg, long sseSrc1Reg, long sseSrc2Reg );
	bool vaddps2(long sseDestReg, long sseSrc1Reg, long sseSrc2Reg);

	bool vaddpd ( long sseDestReg, long sseSrc1Reg, long sseSrc2Reg );
	bool vaddpd2(long sseDestReg, long sseSrc1Reg, long sseSrc2Reg);

	bool vaddss ( long sseDestReg, long sseSrc1Reg, long sseSrc2Reg );
	bool vaddsd ( long sseDestReg, long sseSrc1Reg, long sseSrc2Reg );


	bool vaddps ( long sseDestReg, long sseSrc1Reg, long x64BaseReg, long x64IndexReg, long Scale, long Offset );
	bool vaddpd ( long sseDestReg, long sseSrc1Reg, long x64BaseReg, long x64IndexReg, long Scale, long Offset );
	bool vaddss ( long sseDestReg, long sseSrc1Reg, long x64BaseReg, long x64IndexReg, long Scale, long Offset );
	bool vaddsd ( long sseDestReg, long sseSrc1Reg, long x64BaseReg, long x64IndexReg, long Scale, long Offset );

	// andp
	bool andps ( long sseDestReg, long sseSrc1Reg, long sseSrc2Reg );
	bool andpd ( long sseDestReg, long sseSrc1Reg, long sseSrc2Reg );
	
	bool andps ( long sseDestReg, long sseSrc1Reg, long x64BaseReg, long x64IndexReg, long Scale, long Offset );
	bool andpd ( long sseDestReg, long sseSrc1Reg, long x64BaseReg, long x64IndexReg, long Scale, long Offset );

	// andnp
	bool andnps ( long sseDestReg, long sseSrc1Reg, long sseSrc2Reg );
	bool andnpd ( long sseDestReg, long sseSrc1Reg, long sseSrc2Reg );
	
	bool andnps ( long sseDestReg, long sseSrc1Reg, long x64BaseReg, long x64IndexReg, long Scale, long Offset );
	bool andnpd ( long sseDestReg, long sseSrc1Reg, long x64BaseReg, long x64IndexReg, long Scale, long Offset );

	// cmpxxp

	// avx512
	bool cmppseq_rrr128(long kdst, long src0, long src1, long mask = 0);
	bool cmppsne_rrr128(long kdst, long src0, long src1, long mask = 0);
	bool cmppsgt_rrr128(long kdst, long src0, long src1, long mask = 0);
	bool cmppsge_rrr128(long kdst, long src0, long src1, long mask = 0);
	bool cmppslt_rrr128(long kdst, long src0, long src1, long mask = 0);
	bool cmppsle_rrr128(long kdst, long src0, long src1, long mask = 0);

	bool cmpsseq_rrr128(long kdst, long src0, long src1, long mask = 0);
	bool cmpssne_rrr128(long kdst, long src0, long src1, long mask = 0);
	bool cmpssgt_rrr128(long kdst, long src0, long src1, long mask = 0);
	bool cmpssge_rrr128(long kdst, long src0, long src1, long mask = 0);
	bool cmpsslt_rrr128(long kdst, long src0, long src1, long mask = 0);
	bool cmpssle_rrr128(long kdst, long src0, long src1, long mask = 0);


	bool cmpsdeq_rrr128(long kdst, long src0, long src1, long mask = 0);
	bool cmpsdne_rrr128(long kdst, long src0, long src1, long mask = 0);
	bool cmpsdgt_rrr128(long kdst, long src0, long src1, long mask = 0);
	bool cmpsdge_rrr128(long kdst, long src0, long src1, long mask = 0);
	bool cmpsdlt_rrr128(long kdst, long src0, long src1, long mask = 0);
	bool cmpsdle_rrr128(long kdst, long src0, long src1, long mask = 0);

	bool cmpsdeq_rrm128(long kdst, long src0, void* pSrcPtr, long mask = 0);
	bool cmpsdne_rrm128(long kdst, long src0, void* pSrcPtr, long mask = 0);
	bool cmpsdgt_rrm128(long kdst, long src0, void* pSrcPtr, long mask = 0);
	bool cmpsdge_rrm128(long kdst, long src0, void* pSrcPtr, long mask = 0);
	bool cmpsdlt_rrm128(long kdst, long src0, void* pSrcPtr, long mask = 0);
	bool cmpsdle_rrm128(long kdst, long src0, void* pSrcPtr, long mask = 0);

	bool cmppdeq_rrr128(long kdst, long src0, long src1, long mask = 0);
	bool cmppdne_rrr128(long kdst, long src0, long src1, long mask = 0);
	bool cmppdgt_rrr128(long kdst, long src0, long src1, long mask = 0);
	bool cmppdge_rrr128(long kdst, long src0, long src1, long mask = 0);
	bool cmppdlt_rrr128(long kdst, long src0, long src1, long mask = 0);
	bool cmppdle_rrr128(long kdst, long src0, long src1, long mask = 0);

	bool cmppdeq_rrr256(long kdst, long src0, long src1, long mask = 0);
	bool cmppdne_rrr256(long kdst, long src0, long src1, long mask = 0);
	bool cmppdgt_rrr256(long kdst, long src0, long src1, long mask = 0);
	bool cmppdge_rrr256(long kdst, long src0, long src1, long mask = 0);
	bool cmppdlt_rrr256(long kdst, long src0, long src1, long mask = 0);
	bool cmppdle_rrr256(long kdst, long src0, long src1, long mask = 0);

	bool cmppdeq_rrm128(long kdst, long src0, void* pSrcPtr, long mask = 0);
	bool cmppdne_rrm128(long kdst, long src0, void* pSrcPtr, long mask = 0);
	bool cmppdgt_rrm128(long kdst, long src0, void* pSrcPtr, long mask = 0);
	bool cmppdge_rrm128(long kdst, long src0, void* pSrcPtr, long mask = 0);
	bool cmppdlt_rrm128(long kdst, long src0, void* pSrcPtr, long mask = 0);
	bool cmppdle_rrm128(long kdst, long src0, void* pSrcPtr, long mask = 0);

	// avx2
	bool cmpps ( long sseDestReg, long sseSrc1Reg, long sseSrc2Reg, char Imm8 );
	bool cmppd ( long sseDestReg, long sseSrc1Reg, long sseSrc2Reg, char Imm8 );
	bool cmpss ( long sseDestReg, long sseSrc1Reg, long sseSrc2Reg, char Imm8 );
	bool cmpsd ( long sseDestReg, long sseSrc1Reg, long sseSrc2Reg, char Imm8 );
	bool cmpeqps ( long sseDestReg, long sseSrc1Reg, long sseSrc2Reg );
	bool cmpeqpd ( long sseDestReg, long sseSrc1Reg, long sseSrc2Reg );
	bool cmpeqss ( long sseDestReg, long sseSrc1Reg, long sseSrc2Reg );
	bool cmpeqsd ( long sseDestReg, long sseSrc1Reg, long sseSrc2Reg );
	bool cmpneps(long sseDestReg, long sseSrc1Reg, long sseSrc2Reg);
	bool cmpnepd(long sseDestReg, long sseSrc1Reg, long sseSrc2Reg);
	bool cmpness(long sseDestReg, long sseSrc1Reg, long sseSrc2Reg);
	bool cmpnesd(long sseDestReg, long sseSrc1Reg, long sseSrc2Reg);
	bool cmpltps ( long sseDestReg, long sseSrc1Reg, long sseSrc2Reg );
	bool cmpltpd ( long sseDestReg, long sseSrc1Reg, long sseSrc2Reg );
	bool cmpltss ( long sseDestReg, long sseSrc1Reg, long sseSrc2Reg );
	bool cmpltsd ( long sseDestReg, long sseSrc1Reg, long sseSrc2Reg );
	bool cmpgtps ( long sseDestReg, long sseSrc1Reg, long sseSrc2Reg );
	bool cmpgtpd ( long sseDestReg, long sseSrc1Reg, long sseSrc2Reg );
	bool cmpgtss ( long sseDestReg, long sseSrc1Reg, long sseSrc2Reg );
	bool cmpgtsd ( long sseDestReg, long sseSrc1Reg, long sseSrc2Reg );
	bool cmpleps ( long sseDestReg, long sseSrc1Reg, long sseSrc2Reg );
	bool cmplepd ( long sseDestReg, long sseSrc1Reg, long sseSrc2Reg );
	bool cmpless ( long sseDestReg, long sseSrc1Reg, long sseSrc2Reg );
	bool cmplesd ( long sseDestReg, long sseSrc1Reg, long sseSrc2Reg );
	bool cmpgeps ( long sseDestReg, long sseSrc1Reg, long sseSrc2Reg );
	bool cmpgepd ( long sseDestReg, long sseSrc1Reg, long sseSrc2Reg );
	bool cmpgess ( long sseDestReg, long sseSrc1Reg, long sseSrc2Reg );
	bool cmpgesd ( long sseDestReg, long sseSrc1Reg, long sseSrc2Reg );
	bool cmpunordps ( long sseDestReg, long sseSrc1Reg, long sseSrc2Reg );
	bool cmpunordpd ( long sseDestReg, long sseSrc1Reg, long sseSrc2Reg );
	bool cmpunordss ( long sseDestReg, long sseSrc1Reg, long sseSrc2Reg );
	bool cmpunordsd ( long sseDestReg, long sseSrc1Reg, long sseSrc2Reg );
	bool cmpordps ( long sseDestReg, long sseSrc1Reg, long sseSrc2Reg );
	bool cmpordpd ( long sseDestReg, long sseSrc1Reg, long sseSrc2Reg );
	bool cmpordss ( long sseDestReg, long sseSrc1Reg, long sseSrc2Reg );
	bool cmpordsd ( long sseDestReg, long sseSrc1Reg, long sseSrc2Reg );
	
	bool cmpps ( long sseDestReg, long sseSrc1Reg, long x64BaseReg, long x64IndexReg, long Scale, long Offset, char Imm8 );
	bool cmppd ( long sseDestReg, long sseSrc1Reg, long x64BaseReg, long x64IndexReg, long Scale, long Offset, char Imm8 );
	bool cmpss ( long sseDestReg, long sseSrc1Reg, long x64BaseReg, long x64IndexReg, long Scale, long Offset, char Imm8 );
	bool cmpsd ( long sseDestReg, long sseSrc1Reg, long x64BaseReg, long x64IndexReg, long Scale, long Offset, char Imm8 );
	bool cmpeqps ( long sseDestReg, long sseSrc1Reg, long x64BaseReg, long x64IndexReg, long Scale, long Offset );
	bool cmpeqpd ( long sseDestReg, long sseSrc1Reg, long x64BaseReg, long x64IndexReg, long Scale, long Offset );
	bool cmpeqss ( long sseDestReg, long sseSrc1Reg, long x64BaseReg, long x64IndexReg, long Scale, long Offset );
	bool cmpeqsd ( long sseDestReg, long sseSrc1Reg, long x64BaseReg, long x64IndexReg, long Scale, long Offset );
	bool cmpltps ( long sseDestReg, long sseSrc1Reg, long x64BaseReg, long x64IndexReg, long Scale, long Offset );
	bool cmpltpd ( long sseDestReg, long sseSrc1Reg, long x64BaseReg, long x64IndexReg, long Scale, long Offset );
	bool cmpltss ( long sseDestReg, long sseSrc1Reg, long x64BaseReg, long x64IndexReg, long Scale, long Offset );
	bool cmpltsd ( long sseDestReg, long sseSrc1Reg, long x64BaseReg, long x64IndexReg, long Scale, long Offset );
	bool cmpgtps ( long sseDestReg, long sseSrc1Reg, long x64BaseReg, long x64IndexReg, long Scale, long Offset );
	bool cmpgtpd ( long sseDestReg, long sseSrc1Reg, long x64BaseReg, long x64IndexReg, long Scale, long Offset );
	bool cmpgtss ( long sseDestReg, long sseSrc1Reg, long x64BaseReg, long x64IndexReg, long Scale, long Offset );
	bool cmpgtsd ( long sseDestReg, long sseSrc1Reg, long x64BaseReg, long x64IndexReg, long Scale, long Offset );
	bool cmpleps ( long sseDestReg, long sseSrc1Reg, long x64BaseReg, long x64IndexReg, long Scale, long Offset );
	bool cmplepd ( long sseDestReg, long sseSrc1Reg, long x64BaseReg, long x64IndexReg, long Scale, long Offset );
	bool cmpless ( long sseDestReg, long sseSrc1Reg, long x64BaseReg, long x64IndexReg, long Scale, long Offset );
	bool cmplesd ( long sseDestReg, long sseSrc1Reg, long x64BaseReg, long x64IndexReg, long Scale, long Offset );
	bool cmpgeps ( long sseDestReg, long sseSrc1Reg, long x64BaseReg, long x64IndexReg, long Scale, long Offset );
	bool cmpgepd ( long sseDestReg, long sseSrc1Reg, long x64BaseReg, long x64IndexReg, long Scale, long Offset );
	bool cmpgess ( long sseDestReg, long sseSrc1Reg, long x64BaseReg, long x64IndexReg, long Scale, long Offset );
	bool cmpgesd ( long sseDestReg, long sseSrc1Reg, long x64BaseReg, long x64IndexReg, long Scale, long Offset );
	bool cmpunordps ( long sseDestReg, long sseSrc1Reg, long x64BaseReg, long x64IndexReg, long Scale, long Offset );
	bool cmpunordpd ( long sseDestReg, long sseSrc1Reg, long x64BaseReg, long x64IndexReg, long Scale, long Offset );
	bool cmpunordss ( long sseDestReg, long sseSrc1Reg, long x64BaseReg, long x64IndexReg, long Scale, long Offset );
	bool cmpunordsd ( long sseDestReg, long sseSrc1Reg, long x64BaseReg, long x64IndexReg, long Scale, long Offset );
	bool cmpordps ( long sseDestReg, long sseSrc1Reg, long x64BaseReg, long x64IndexReg, long Scale, long Offset );
	bool cmpordpd ( long sseDestReg, long sseSrc1Reg, long x64BaseReg, long x64IndexReg, long Scale, long Offset );
	bool cmpordss ( long sseDestReg, long sseSrc1Reg, long x64BaseReg, long x64IndexReg, long Scale, long Offset );
	bool cmpordsd ( long sseDestReg, long sseSrc1Reg, long x64BaseReg, long x64IndexReg, long Scale, long Offset );
	
	// cvtp

	bool cvtdq2pd ( long sseDestReg, long sseSrc1Reg );
	bool cvtdq2ps ( long sseDestReg, long sseSrc1Reg );
	bool cvtpd2dq ( long sseDestReg, long sseSrc1Reg );
	bool cvtps2dq ( long sseDestReg, long sseSrc1Reg );
	bool cvtps2pd ( long sseDestReg, long sseSrc1Reg );
	bool cvtpd2ps ( long sseDestReg, long sseSrc1Reg );
	
	bool cvtdq2pd ( long sseDestReg, long x64BaseReg, long x64IndexReg, long Scale, long Offset );
	bool cvtdq2ps ( long sseDestReg, long x64BaseReg, long x64IndexReg, long Scale, long Offset );
	bool cvtpd2dq ( long sseDestReg, long x64BaseReg, long x64IndexReg, long Scale, long Offset );
	bool cvtps2dq ( long sseDestReg, long x64BaseReg, long x64IndexReg, long Scale, long Offset );
	bool cvtps2pd ( long sseDestReg, long x64BaseReg, long x64IndexReg, long Scale, long Offset );
	bool cvtpd2ps ( long sseDestReg, long x64BaseReg, long x64IndexReg, long Scale, long Offset );

	// avx 2

	bool cvtpd2ps1regreg(long sseDestReg, long sseSrc1Reg);
	bool cvtpd2ps2regreg(long sseDestReg, long sseSrc1Reg);

	bool cvtps2pd1regreg(long sseDestReg, long sseSrc1Reg);
	bool cvtps2pd2regreg(long sseDestReg, long sseSrc1Reg);

	// divp
	bool vdivps ( long sseDestReg, long sseSrc1Reg, long sseSrc2Reg );
	bool vdivpd ( long sseDestReg, long sseSrc1Reg, long sseSrc2Reg );
	bool vdivss ( long sseDestReg, long sseSrc1Reg, long sseSrc2Reg );
	bool vdivsd ( long sseDestReg, long sseSrc1Reg, long sseSrc2Reg );
	
	bool vdivps ( long sseDestReg, long sseSrc1Reg, long x64BaseReg, long x64IndexReg, long Scale, long Offset );
	bool vdivpd ( long sseDestReg, long sseSrc1Reg, long x64BaseReg, long x64IndexReg, long Scale, long Offset );
	bool vdivss ( long sseDestReg, long sseSrc1Reg, long x64BaseReg, long x64IndexReg, long Scale, long Offset );
	bool vdivsd ( long sseDestReg, long sseSrc1Reg, long x64BaseReg, long x64IndexReg, long Scale, long Offset );

	//maxp
	bool maxps ( long sseDestReg, long sseSrc1Reg, long sseSrc2Reg );
	bool maxpd ( long sseDestReg, long sseSrc1Reg, long sseSrc2Reg );
	bool maxss ( long sseDestReg, long sseSrc1Reg, long sseSrc2Reg );
	bool maxsd ( long sseDestReg, long sseSrc1Reg, long sseSrc2Reg );
	
	bool maxps ( long sseDestReg, long sseSrc1Reg, long x64BaseReg, long x64IndexReg, long Scale, long Offset );
	bool maxpd ( long sseDestReg, long sseSrc1Reg, long x64BaseReg, long x64IndexReg, long Scale, long Offset );
	bool maxss ( long sseDestReg, long sseSrc1Reg, long x64BaseReg, long x64IndexReg, long Scale, long Offset );
	bool maxsd ( long sseDestReg, long sseSrc1Reg, long x64BaseReg, long x64IndexReg, long Scale, long Offset );

	// avx2
	bool vmaxsd(long sseDestReg, long sseSrc1Reg, long sseSrc2Reg);
	bool vmaxpd(long sseDestReg, long sseSrc1Reg, long sseSrc2Reg);
	bool vmaxpd2(long sseDestReg, long sseSrc1Reg, long sseSrc2Reg);

	bool vmaxps1_regmem(long sseDestReg, long sseSrc1Reg, void* pSrcPtr);
	bool vmaxpd1_regmem(long sseDestReg, long sseSrc1Reg, void* pSrcPtr);
	bool vmaxss1_regmem(long sseDestReg, long sseSrc1Reg, void* pSrcPtr);
	bool vmaxsd1_regmem(long sseDestReg, long sseSrc1Reg, void* pSrcPtr);

	// minp
	bool minps ( long sseDestReg, long sseSrc1Reg, long sseSrc2Reg );
	bool minpd ( long sseDestReg, long sseSrc1Reg, long sseSrc2Reg );
	bool minss ( long sseDestReg, long sseSrc1Reg, long sseSrc2Reg );
	bool minsd ( long sseDestReg, long sseSrc1Reg, long sseSrc2Reg );
	
	bool minps ( long sseDestReg, long sseSrc1Reg, long x64BaseReg, long x64IndexReg, long Scale, long Offset );
	bool minpd ( long sseDestReg, long sseSrc1Reg, long x64BaseReg, long x64IndexReg, long Scale, long Offset );
	bool minss ( long sseDestReg, long sseSrc1Reg, long x64BaseReg, long x64IndexReg, long Scale, long Offset );
	bool minsd ( long sseDestReg, long sseSrc1Reg, long x64BaseReg, long x64IndexReg, long Scale, long Offset );

	// avx2
	bool vminsd(long sseDestReg, long sseSrc1Reg, long sseSrc2Reg);
	bool vminpd(long sseDestReg, long sseSrc1Reg, long sseSrc2Reg);
	bool vminpd2(long sseDestReg, long sseSrc1Reg, long sseSrc2Reg);

	bool vminps1_regmem(long sseDestReg, long sseSrc1Reg, void* pSrcPtr);
	bool vminpd1_regmem(long sseDestReg, long sseSrc1Reg, void* pSrcPtr);
	bool vminss1_regmem(long sseDestReg, long sseSrc1Reg, void* pSrcPtr);
	bool vminsd1_regmem(long sseDestReg, long sseSrc1Reg, void* pSrcPtr);


	// range

	bool vrangesd_rri64(long dst, long src0, long src1, char Imm8, long mask = 0, long z = 0);
	bool vrangesd_rmi64(long dst, long src0, void* pSrcPtr64, char Imm8, long mask = 0, long z = 0);

	bool vrangess_rri32(long dst, long src0, long src1, char Imm8, long mask = 0, long z = 0);
	bool vrangess_rmi32(long dst, long src0, void* pSrcPtr32, char Imm8, long mask = 0, long z = 0);

	bool vrangeps_rri128(long dst, long src0, long src1, char Imm8, long mask = 0, long z = 0);
	bool vrangeps_rri256(long dst, long src0, long src1, char Imm8, long mask = 0, long z = 0);
	bool vrangeps_rri512(long dst, long src0, long src1, char Imm8, long mask = 0, long z = 0);

	bool vrangeps_rmi128(long dst, long src0, void* pSrcPtr64, char Imm8, long mask = 0, long z = 0);
	bool vrangeps_rmi256(long dst, long src0, void* pSrcPtr64, char Imm8, long mask = 0, long z = 0);
	bool vrangeps_rmi512(long dst, long src0, void* pSrcPtr64, char Imm8, long mask = 0, long z = 0);

	bool vrangepd_rri128(long dst, long src0, long src1, char Imm8, long mask = 0, long z = 0);
	bool vrangepd_rri256(long dst, long src0, long src1, char Imm8, long mask = 0, long z = 0);
	bool vrangepd_rri512(long dst, long src0, long src1, char Imm8, long mask = 0, long z = 0);

	bool vrangepd_rmi128(long dst, long src0, void* pSrcPtr64, char Imm8, long mask = 0, long z = 0);
	bool vrangepd_rmi256(long dst, long src0, void* pSrcPtr64, char Imm8, long mask = 0, long z = 0);
	bool vrangepd_rmi512(long dst, long src0, void* pSrcPtr64, char Imm8, long mask = 0, long z = 0);

	bool vrangepd_rbi128(long dst, long src0, long long* pSrcPtr64, char Imm8, long mask = 0, long z = 0);
	bool vrangepd_rbi256(long dst, long src0, long long* pSrcPtr64, char Imm8, long mask = 0, long z = 0);
	bool vrangepd_rbi512(long dst, long src0, long long* pSrcPtr64, char Imm8, long mask = 0, long z = 0);


	// mulp

	// avx512
	bool vmulpd_rrr512(long dst, long src0, long src1, long mask = 0, long z = 0);
	bool vmulpd_rrr256(long dst, long src0, long src1, long mask = 0, long z = 0);
	bool vmulpd_rrr128(long dst, long src0, long src1, long mask = 0, long z = 0);

	bool vmulps_rrr512(long dst, long src0, long src1, long mask = 0, long z = 0);
	bool vmulps_rrr256(long dst, long src0, long src1, long mask = 0, long z = 0);
	bool vmulps_rrr128(long dst, long src0, long src1, long mask = 0, long z = 0);

	bool vmulss_rrr128(long dst, long src0, long src1, long mask = 0, long z = 0);
	bool vmulsd_rrr128(long dst, long src0, long src1, long mask = 0, long z = 0);

	// avx2
	bool vmulps ( long sseDestReg, long sseSrc1Reg, long sseSrc2Reg );
	bool vmulps2(long sseDestReg, long sseSrc1Reg, long sseSrc2Reg);

	bool vmulpd ( long sseDestReg, long sseSrc1Reg, long sseSrc2Reg );
	bool vmulpd2(long sseDestReg, long sseSrc1Reg, long sseSrc2Reg);

	bool vmulss ( long sseDestReg, long sseSrc1Reg, long sseSrc2Reg );
	bool vmulsd ( long sseDestReg, long sseSrc1Reg, long sseSrc2Reg );
	
	bool vmulss1_regmem(long sseDestReg, long sseSrc1Reg, void* pSrcPtr);
	bool vmulsd1_regmem(long sseDestReg, long sseSrc1Reg, void* pSrcPtr);

	bool vmulps ( long sseDestReg, long sseSrc1Reg, long x64BaseReg, long x64IndexReg, long Scale, long Offset );
	bool vmulpd ( long sseDestReg, long sseSrc1Reg, long x64BaseReg, long x64IndexReg, long Scale, long Offset );
	bool vmulss ( long sseDestReg, long sseSrc1Reg, long x64BaseReg, long x64IndexReg, long Scale, long Offset );
	bool vmulsd ( long sseDestReg, long sseSrc1Reg, long x64BaseReg, long x64IndexReg, long Scale, long Offset );


	// orp

	// avx2
	bool vorps ( long sseDestReg, long sseSrc1Reg, long sseSrc2Reg );
	bool vorpd ( long sseDestReg, long sseSrc1Reg, long sseSrc2Reg );

	bool vorps ( long sseDestReg, long sseSrc1Reg, long x64BaseReg, long x64IndexReg, long Scale, long Offset );
	bool vorpd ( long sseDestReg, long sseSrc1Reg, long x64BaseReg, long x64IndexReg, long Scale, long Offset );


	// shufp

	// avx2
	bool vshufps ( long sseDestReg, long sseSrc1Reg, long sseSrc2Reg, char Imm8 );
	bool vshufpd ( long sseDestReg, long sseSrc1Reg, long sseSrc2Reg, char Imm8 );

	bool vshufps ( long sseDestReg, long sseSrc1Reg, long x64BaseReg, long x64IndexReg, long Scale, long Offset, char Imm8 );
	bool vshufpd ( long sseDestReg, long sseSrc1Reg, long x64BaseReg, long x64IndexReg, long Scale, long Offset, char Imm8 );
	
	// rsqrtp - I don't think this is used
	
	// sqrtp

	// avx2
	bool vsqrtps ( long sseDestReg, long sseSrc1Reg );
	bool vsqrtpd ( long sseDestReg, long sseSrc1Reg );
	bool vsqrtss ( long sseDestReg, long sseSrc1Reg );
	bool vsqrtsd ( long sseDestReg, long sseSrc1Reg );
	
	bool vsqrtps ( long sseDestReg, long x64BaseReg, long x64IndexReg, long Scale, long Offset );
	bool vsqrtpd ( long sseDestReg, long x64BaseReg, long x64IndexReg, long Scale, long Offset );
	bool vsqrtss ( long sseDestReg, long x64BaseReg, long x64IndexReg, long Scale, long Offset );
	bool vsqrtsd ( long sseDestReg, long x64BaseReg, long x64IndexReg, long Scale, long Offset );
	
	// subp

	// sse
	bool subpsregreg ( long sseDestReg, long sseSrcReg );
	bool subpdregreg ( long sseDestReg, long sseSrcReg );
	
	// avx2
	bool vsubps ( long sseDestReg, long sseSrc1Reg, long sseSrc2Reg );
	bool vsubpd ( long sseDestReg, long sseSrc1Reg, long sseSrc2Reg );
	bool vsubss ( long sseDestReg, long sseSrc1Reg, long sseSrc2Reg );
	bool vsubsd ( long sseDestReg, long sseSrc1Reg, long sseSrc2Reg );
	
	bool vsubps2(long sseDestReg, long sseSrc1Reg, long sseSrc2Reg);
	bool vsubpd2(long sseDestReg, long sseSrc1Reg, long sseSrc2Reg);

	bool vsubps ( long sseDestReg, long sseSrc1Reg, long x64BaseReg, long x64IndexReg, long Scale, long Offset );
	bool vsubpd ( long sseDestReg, long sseSrc1Reg, long x64BaseReg, long x64IndexReg, long Scale, long Offset );
	bool vsubss ( long sseDestReg, long sseSrc1Reg, long x64BaseReg, long x64IndexReg, long Scale, long Offset );
	bool vsubsd ( long sseDestReg, long sseSrc1Reg, long x64BaseReg, long x64IndexReg, long Scale, long Offset );
	
	// avx-512
	bool vsubps_rrr512(long dst, long src0, long src1, long mask = 0, long z = 0, long rnd_enable = 0, long rnd_type = 0);
	bool vsubps_rrr256(long dst, long src0, long src1, long mask = 0, long z = 0);
	bool vsubps_rrr128(long dst, long src0, long src1, long mask = 0, long z = 0);

	bool vsubpd_rrr512(long dst, long src0, long src1, long mask = 0, long z = 0);
	bool vsubpd_rrr256(long dst, long src0, long src1, long mask = 0, long z = 0);
	bool vsubpd_rrr128(long dst, long src0, long src1, long mask = 0, long z = 0);

	bool vsubss_rrr128(long dst, long src0, long src1, long mask = 0, long z = 0, long rnd_enable = 0, long rnd_type = 0);

	bool vsubsd_rrr128(long dst, long src0, long src1, long mask = 0, long z = 0);

	// vperm2f128

	// avx2
	bool vperm2f128 ( long sseDestReg, long sseSrc1Reg, long sseSrc2Reg, char Imm8 );
	
	// xorp

	// avx2
	bool vxorps ( long sseDestReg, long sseSrc1Reg, long sseSrc2Reg );
	bool vxorpd ( long sseDestReg, long sseSrc1Reg, long sseSrc2Reg );
	
	bool vxorps ( long sseDestReg, long sseSrc1Reg, long x64BaseReg, long x64IndexReg, long Scale, long Offset );
	bool vxorpd ( long sseDestReg, long sseSrc1Reg, long x64BaseReg, long x64IndexReg, long Scale, long Offset );
	
	
	// **** bookmarking **** //
	
	// if you have a jump and don't know how far you'll need to jump, you'll need to store the jump amount later
	// this will return a bookmark for the current byte and advance to the next byte so you can insert the next instruction
	long x64GetBookmark8 ( void );
	
	// this is where you can write to where you put the bookmark - this will be used for short jumps
	bool x64SetBookmark8 ( long Bookmark, char value );

private:

	bool x64EncodeReg16 ( long x64InstOpcode, long ModRMOpcode, long x64Reg );

	// general encode of sib and immediates and stuff for register-memory instructions
	inline bool x64EncodeMem ( long x64DestReg, long BaseAddressReg, long IndexReg, long Scale, long Offset );
	
	inline bool x64Encode16Bit ( void );
	inline bool x64EncodePrefix ( char Prefix );
	
	bool x64EncodeRexReg32 ( long DestReg_RM_Base, long SourceReg_Reg_Opcode );
	bool x64EncodeRexReg64 ( long DestReg_RM_Base, long SourceReg_Reg_Opcode );

	// general encoding of opcode(s)
	inline bool x64EncodeOpcode ( long x64InstOpcode );
};

#endif	// end #ifndef __X64ENCODER_H__
